#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 14:05:10 2018

@author: yoelr
"""
from . import _Q
import copy
import numpy as np
from scipy.optimize import newton, least_squares
from ._utils import material_array, property_array, PropertyFactory, \
                    tuple_array, fraction, Sink, Source
from ._flowsheet import find
from ._species import Species
from .exceptions import SolverError, EquilibriumError, DimensionError
from ._equilibrium import DORTMUND

inf = np.inf
ln = np.log
exp = np.exp
array = np.array

__all__ = ('Stream',)


# %% TODOs

# TODO: add material property interphase when using Cape Open package


# %% Functions

def nonzero_species(num_IDs, flow):
    index = []
    species = []
    for i, ID in num_IDs:
        if flow[i] != 0:
            index.append(i)
            species.append(ID)
    return index, species

def _print_helpdata(helpdata):
    """Print help data."""
    # Only one helplist, print a nice string
    if isinstance(helpdata[0], str):
        propID, description, dependency, units, datatype = helpdata
        if dependency == 'TP':
            dependency = 'as a function of T and P '
        elif dependency == 'T':
            dependency = 'as a function of T '
        elif dependency == 'P':
            dependency = 'as a function of P '
        print(f"{propID}: [{datatype}] {description.capitalize()} {dependency}({units}).")

    # Many helpdata, print all the nice strings
    else:
        for i in helpdata:
            _print_helpdata(i)


# %% Units of measure

# Biosteam units of measure
units_of_measure = dict(MW='g/mol',
                        mass='kg/hr',
                        mol='kmol/hr',
                        vol='m^3/hr',
                        massnet='kg/hr',
                        molnet='kmol/hr',
                        volnet='m^3/hr',
                        massfrac='kg/kg',
                        molfrac='kmol/kmol',
                        volfrac='m^3/m^3',
                        T='K',
                        P='Pa',
                        H='kJ/hr',
                        S='kJ/hr',
                        G='kJ/hr',
                        U='kJ/hr',
                        A='kJ/hr',
                        Hf='kJ/hr',
                        C='kJ/K/hr',
                        Vm='m^3/mol',
                        Cpm='J/mol/K',
                        Cp='J/g/K',
                        rho='kg/m^3',
                        rhom='mol/m^3',
                        nu='m^2/s',
                        mu='Pa*s',
                        sigma='N/m',
                        k='W/m/K',
                        alpha='m^2/s')

mol_flow_dim = _Q(0, units_of_measure['mol']).dimensionality
mass_flow_dim = _Q(0, units_of_measure['mass']).dimensionality
vol_flow_dim = _Q(0, units_of_measure['vol']).dimensionality


# %% Flow properties

@PropertyFactory
def MassFlow(self):
    """Mass flow (kg/hr)."""
    mol, MW = self.data
    m = mol.item(0)
    if m: return MW * m
    else: return 0.

@MassFlow.setter
def MassFlow(self, value):
    mol, MW = self.data
    mol[0] = value/MW if value else 0.

@PropertyFactory    
def VolumetricFlow(self):
    """Volumetric flow (m^3/hr)."""
    stream, mol = self.data
    m = mol[0]
    if m:
        c = self.name # c = compound
        c.T = stream.T
        c.P = stream.P
        c.phase = stream.phase
        return c.Vm * m * 1000
    else:
        return 0.

@VolumetricFlow.setter
def VolumetricFlow(self, value):
    stream, mol = self.data
    if value:
        c = self.name # c = compound
        c.T = stream.T
        c.P = stream.P
        c.phase = stream.phase
        mol[0] = value/(c.Vm * 1000)
    else:
        mol[0] = 0.


# %% Stream classes


class metaStream(type):
    """Metaclass for Stream."""
    @property
    def species(cls):
        """[Species] Contains pure component thermodynamic properties for computing overall properties of Stream instances."""
        return cls._species

    @species.setter
    def species(cls, species):
        # Set Species object and related parameters
        if species is cls._species: pass
        elif isinstance(species, Species):
            tup = tuple
            compounds = tup(species)
            IDs = tup(i.ID for i in compounds)
            CAS = tup(i.CAS for i in compounds)
            Nspecies = len(IDs)
            index = tup(range(Nspecies))
            MW = array([i.MW for i in compounds])
            for cl in (Stream, MS.MixedStream):
                cl._num_compounds = tup(zip(index, compounds))
                cl._num_IDs = tup(zip(index, IDs))
                cl._species_dict = species.__dict__
                cl._compounds = compounds
                cl._species = species
                cl._IDs = IDs
                cl._CAS = CAS
                cl._Nspecies = Nspecies
                cl._MW = MW
            species._immutable.add(species)
        else: raise ValueError('Must pass a Species object')


class Stream(metaclass=metaStream):
    """Create a Stream object that defines material flow rates along its thermodynamic state. Thermodynamic and transport properties of a stream are readily available. Ideal mixture is assumed for stream properties and excess thermodynamic energies are neglected as a simplifying assumption for low pressure processes.

**Parameters**

    **ID:** [str] A unique identification. If ID is an empty string (i.e. '' ), a default ID will be chosen. If ID is None, stream will not be registered in flowsheet.

    **flow:** [tuple] All flow rates corresponding to `species`.

    **species:** tuple[str] or [Species] Species corresponding to `flow`. If empty, assume same species as class.

    **units:** [str] The units of the flow rates (only mass, molar, and volumetric flow rates are valid)

    **phase:** [str] 'l' for liquid, 'g' for gas, or 's' for solid.

    **T:** [float] Temperature (K).

    **P:** [float] Pressure (Pa).

    **price:** [float] (USD/kg).
    
    ****flow_pairs:** Compound-flow pairs,

**Class Properties**

    **species** = Species(): [Species] Contains pure component thermodynamic properties for computing overall properties of Stream instances.

**Examples**

    Before making a stream, set the species using a Species object:

    .. code-block:: python

       >>> # Set Species object
       >>> Stream.species = Species('Ethanol', 'Water') 

    Stream objects may be created a variety of ways:

    .. code-block:: python

       >>> # Create a stream specifying compound and flow rate pairs:
       >>> s1 = Stream(ID='s1', Water=2)
       >>> s1.show()
       Stream: s1
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  2

       >>> # Create a stream assuming same order as given in species:
       >>> s2 = Stream(ID='s2', flow=(1, 2))
       >>> s2.show()
       Stream: s2
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  1
                        Water    2

       >>> # Create a stream passing flow rate units, phase, temperature and pressure:
       >>> s3 = Stream(ID='s3', flow=(0.278, 0.556), units='mol/s', phase='g', T=400, P=101325)
       >>> s3.show()
       Stream: s3
        phase: 'g', T: 400 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  1.
                        Water    2.

       >>> # The working units do not change
       >>> s3.mol 
       material_array([1., 2.])

    .. Warning:: Stream objects do not automatically calculate thermodynamic equilibrium. They simply assume the given phase, temperature and pressure are correct. To find equilibrium, use the VLE or LLE method.

    .. Note:: All following examples will use the Stream objects created here (s1, s2, and s3) starting from its initial values.

    Use the `show` method to print all specifications with desired units:

    .. code-block:: python

       >>> # Temperature in degree Celsius
       >>> s2.show(T='degC')
       Stream: s2 
        phase: 'l', T: 25. degC, P: 101325 Pa
        flow (kmol/hr): Ethanol  1
                        Water    2
       
       >>> # Flow in kg/hr
       >>> s2.show(flow='kg/hr')
       Stream: s2 
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kg/hr): Ethanol  46.1
                      Water    36

       >>> # Flow in fractions
       >>> s2.show(fraction=True)
       Stream: s2
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: Ethanol  0.333
              Water    0.667
              net      3 kmol/hr

    Flow rates are stored internally as a material_array in the ‘mol’ attribute.

    .. code-block:: python

       >>> # Set Water flow rate
       >>> s2.mol[1] = 18
       >>> s2.mol # kmol/hr
       material_array([1, 18])
       
       >>> # A negative flow issues a RuntimeWarning
       >>> s2.mol[1] = -1
       __main__:1: RuntimeWarning:
       Encountered negative or non-finite value in 'material_array' object.
       
       >>> # Setting the property sets its values in place
       >>> s2.mol = [1, 2]
       >>> s2.mol
       material_array([1, 2])
    .. Note::

       material_array objects are numpy ndarrays which issue a RuntimeWarning when a negative or non-finite number is encountered.

    Mass and volumetric flow rates are also available as property_arrays of the molar flow rate. As such, they are always up to date with the molar flow rate and altering them also alters the molar flow rate:

    .. code-block:: python

       >>> # Altering mass or volumetric flows alters the molar flow
       >>> s2.vol # m^3/hr
       property_array([0.059, 0.036])
       >>> s2.vol[:] = [1, 1]
       >>> s2.mol
       material_array([17.06 , 55.343])
       >>> # Values are always up to date with the molar flow
       >>> s2.mol[:] = [1, 2]
       >>> s2.vol
       property_array([0.059, 0.036])

    .. Note::

       property_array objects are significantly slower than material_array objects. This is because flow rate data is internally stored as molar flow rates. Also, property_array objects are arrays of python objects, which add overhead over the C implementation of numpy. Whenever possible, use the material_array to manage flow rates. 

    Some thermodynamic/material properties, including enthalpy and heat capacity, are dependent only on temperature:

    .. code-block:: python
    
       >>> # Enthalpy (kJ/hr) and heat capacity (kJ/(kg-K))
       >>> s2.T = 298.15
       >>> s2.Cp
       3.200382054794244
       >>> s2.H 
       0.0
       
       >>> # Change of temperature
       >>> s2.T = 310
       >>> s2.H
       3140.9389548625936
       >>> s2.Cp
       3.2590872180092956
       
    .. Note:: Thermodynamic energies are relative to 25 degC and 1 atm.
    
    Some thermodynamic/material properties, including volume and density, are dependent on both temperature and pressure:

    .. code-block:: python
    
       >>> # Volumetric flow rate (m^3/hr) and density (kg/m^3)
       >>> s1.volnet
       0.036138079740245625
       >>> s1.rho
       997.0247522552814
      
       >>> # Change of pressure
       >>> s1.P *= 4
       >>> s1.volnet
       0.036136231155141196
       >>> s1.rho
       997.0757560552587
      
       >>> # Change of temperature
       >>> s1.T += 30
       >>> s1.volnet
       0.03663597700908213
       >>> s1.rho
       983.4747955832584

    A dictionary of available stream properties is available in `Stream.units`. You may also find it useful to use the `help` method to search for a property:
        
    .. code-block:: python

       >>> Stream.help('conductivity')
       k: [float] Thermal conductivity as a function of T and P (W/m/K).

    """
    activity_coefficients = staticmethod(DORTMUND)

    # [dict] Units of measure for material properties (class attribute). 
    units = units_of_measure

    # Information regarding properties
    _prop_info = (
        # ID         # Description               # Dependency # Units      # Type
        ('T',        'temperature',              '',          'K',         'float'),
        ('H',        'enthalpy',                 'T',         'kJ/hr',     'float'),
        ('S',        'entropy',                  'TP',        'kJ/hr',     'float'),
        ('G',        'Gibbs free energy',        'TP',        'kJ/hr',     'float'),
        ('U',        'interal energy',           'TP',        'kJ/hr',     'float'),
        ('A',        'Helmholtz free energy',    'TP',        'kJ/hr',     'float'),
        ('Hf',       'enthalpy of formation',    '',          'kJ/hr',     'float'),
        ('P',        'pressure',                 '',          'Pa',        'float'),
        ('Cpm',      'molar heat capacity',      'T',         'J/mol/K',   'float'),
        ('Cp',       'specific heat capacity',   'T',         'J/kg/K',    'float'),
        ('Vm',       'molar volume',             'TP',        'm^3/mol',   'float'),
        ('rho',      'density',                  'TP',        'kg/m^3',    'float'),
        ('rhom',     'molar density',            'TP',        'mol/m^3',   'float'),
        ('nu',       'kinematic viscosity',      'TP',        'm^2/s',     'float'),
        ('mu',       'hydraulic viscosity',      'TP',        'Pa*s',      'float'),
        ('sigma',    'surface tension',          'T',         'N/m',       'float'),
        ('k',        'thermal conductivity',     'TP',        'W/m/K',     'float'),
        ('alpha',    'thermal diffusivity',      'TP',        'm^2/s',     'float'),
        ('Pr',       'Prantl number',            'TP',        "''",        'float'),
        ('mass',     'mass flow rates',          '',          'kg/hr',     'ndarray'),
        ('mol',      'molar flow rates',         '',          'kmol/hr',   'ndarray'),
        ('vol',      'volumetric flow rates',    'TP',        'm^3/hr',    'ndarray'),
        ('massnet',  'net mass flow rate',       '',          'kg/hr',     'float'),
        ('molnet',   'net molar flow rate',      '',          'kmol/hr',   'float'),
        ('volnet',   'net volumetric flow rate', 'TP',        'm^3/hr',    'float'),
        ('massfrac', 'mass fractions',           '',          'kg/kg',     'ndarray'),
        ('molfrac',  'molar fractions',          '',          'kmol/kmol', 'ndarray'),
        ('volfrac',  'volumetric fractions',     'TP',        'm^3/m^3',   'ndarray'))

    ### Class attributes for working species ###
    
    #: Species defined in stream
    _species = None
    
    #: CAS of species
    _CAS = ()

    #: Compounds of species
    _compounds = _CAS
    
    #: Enumerated compounds
    _num_compounds = _CAS
    
    #: Species dictionary
    _species_dict = {}
    
    #: Enumerated species IDs
    _num_IDs = _CAS
    
    #: Species IDs
    _IDs = _CAS
    
    #: Number of species
    _Nspecies = 0  
    
    #: Array of molecular weights
    _MW = _CAS

    # [list] Default starting letter and current number for ID (class attribute)
    _default_ID = ['d', 1]

    # Default ID
    _ID = None
    
    #: [bool] If True, approximate energy balance. False otherwise.
    lazy_energy_balance = True

    line = 'Stream'

    def __init__(self, ID='', flow=(), species=(), units='kmol/hr',
                 phase='l', T=298.15, P=101325, *, price=0, **flow_pairs):
        # Get species and set species information
        if isinstance(species, Species): self._set_species(species)
        else: self._copy_species(self)
        self.ID = ID
        
        #: Dew point cached
        # {species: (pressure,
        #            temperature,
        #            vapor composition,
        #            liquid composition)}
        self._dew_cached = {}
        
        
        #: Bubble point cached
        # {species: (pressure,
        #            temperature,
        #            vapor composition,
        #            liquid composition)}
        self._bubble_cached = {}
        
        #: Vapor composition cached
        # {species: y}
        self._y_cached = {}
        
        #: liquid-LIQUID split cached
        # {species: lL_split}
        self._lL_split_cached = {}
        
        #: Unit source
        self._source = None
        
        #: Unit sink
        self._sink = None
        
        #: [str] 'l' for liquid, 'g' for gas, or 's' for solid.
        self.phase = phase
        
        #: [float] Temperature (K)
        self.T = T
        
        #: [float] Pressure (Pa)
        self.P = P  
        
        # Initialize flows
        self._init_molarray(flow, species, flow_pairs)
        self._mol = mol = self._molarray.view(material_array)
        MW = self._MW
        mass = [] # Mass flow rates
        vol = [] # Volumetric flow rates    
        for i, s in self._num_compounds:
            mol_i = mol[i:i+1]
            m = MassFlow(s.ID, (mol_i, MW[i]))
            v = VolumetricFlow(s, (self, mol_i))
            mass.append(m); vol.append(v)
        self._mass = property_array(mass)
        self._vol = property_array(vol)
        if units != 'kmol/hr': self._set_flowarray_with_units(mol, units)

        #: Price of stream (USD/kg)
        self.price = price

    def setflow(self, flow=(), species=(), units='kmol/hr', inplace='', **flow_pairs):
        """Set `flow` rates according to the `species` order and `flow_pairs`. `inplace` can be any operation that can be performed in place (e.g. +, -, *, /, |, **, etc.)."""
        species = (*species, *flow_pairs.keys())
        flow = (*flow, *flow_pairs.values())
        index = self.indices(*species) if species else ... 
        if units == 'kmol/hr': exec(f"self._mol[index] {inplace}= flow", locals())
        elif units == 'kg/hr': exec(f"self._mass[index] {inplace}= flow", locals())
        elif units == 'm3/hr': exec(f"self._vol[index] {inplace}= flow", locals())
        else:
            flow_wt_units = _Q(flow, units)
            dim = flow_wt_units.dimensionality
            if dim == mol_flow_dim:
                exec(f"self._mol[index] {inplace}= flow_wt_units.to('kmol/hr').magnitude", locals())
            elif dim == mass_flow_dim:
                exec(f"self._mass[index] {inplace}= flow_wt_units.to('kg/hr').magnitude", locals())
            elif dim == vol_flow_dim:
                exec(f"self._vol[index] {inplace}= flow_wt_units.to('m3/hr').magnitude", locals())
            else:
                raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'.")
        
    
    def getflow(self, *species, units='kmol/hr'):
        """Get flow rates of species in given units."""
        index = self.indices(*species) if species else ...
        if units=='kmol/hr': return self._mol[index]    
        flow_wt_units = _Q(1, units)
        dim = flow_wt_units.dimensionality
        if dim == mol_flow_dim:
            return self._molarray[index]*flow_wt_units.to('kmol/hr').magnitude
        elif dim == mass_flow_dim:
            return self._mass[index]*flow_wt_units.to('kg/hr').magnitude
        elif dim == vol_flow_dim:
            return self._vol[index]*flow_wt_units.to('m3/hr').magnitude
        else:
            raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'.")
    
    def _init_molarray(self, flow, species, flow_pairs):
        """Initialize molar flow rates accoring the species order, and flow_pairs. Instance species do not change."""
        lenflow = len(flow)
        empty_flow = lenflow == 0
        if species is self._species:
            if flow_pairs:
                raise ValueError('Cannot specify flow pairs when a Species object is passed')
            elif empty_flow:
                self._molarray = np.zeros(self._Nspecies, float)
            elif lenflow == self._Nspecies:
                self._molarray = np.array(flow, float)
            else:
                raise ValueError('If flow is not empty, length of flow must be equal to length of Species object.')
        else:
            lenspecies = len(species)
            empty_species = lenspecies == 0
            self._molarray = np.zeros(self._Nspecies, float)
            if flow_pairs and empty_flow and empty_species:
                species = flow_pairs.keys()
                self._molarray[self.indices(*species)] = (*flow_pairs.values(),)
            elif lenspecies == lenflow:
                if flow_pairs:
                    self._molarray[self.indices(*species, *flow_pairs.keys())] = (*flow, *flow_pairs.values())
                elif not empty_flow:
                    self._molarray[self.indices(*species)] = flow
            elif empty_species and lenflow == self._Nspecies:
                self._molarray = np.array(flow, float)
            else:
                raise ValueError('Length of flow rates must be equal to length of species')
    
    def _set_flowarray_with_units(self, flow, units):
        """Set flow rates accorthing to given units."""
        if units == 'kg/hr': self._mass[:] = flow
        elif units == 'm3/hr': self._vol[:] = flow
        else:
            flow_wt_units = _Q(flow, units)
            dim = flow_wt_units.dimensionality
            if dim == mol_flow_dim:
                self._mol[:] = flow_wt_units.to('kmol/hr').magnitude
            elif dim == mass_flow_dim:
                self._mass[:] = flow_wt_units.to('kg/hr').magnitude
            elif dim == vol_flow_dim:
                self._vol[:] = flow_wt_units.to('m3/hr').magnitude
            else:
                raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'.")

    def _set_species(self, species):
        """Set Species object and related information."""
        if self._species is species: self._copy_species(self)
        species._immutable.add(species)
        tup = tuple
        self._species = species
        self._compounds = compounds = tup(species)
        self._CAS = tup(i.CAS for i in compounds)
        self._IDs = IDs = tup(i.ID for i in compounds)
        self._Nspecies = len(IDs)  
        index = tup(range(self._Nspecies))
        self._species_dict = species.__dict__
        self._num_compounds = tup(zip(index, compounds))
        self._num_IDs = tup(zip(index, IDs))
        self._MW = array([c.MW for c in compounds])

    def _copy_species(self, stream):
        """Copy species data from stream."""
        self._compounds = stream._compounds
        self._num_compounds = stream._num_compounds
        self._num_IDs = stream._num_IDs
        self._Nspecies = stream._Nspecies
        self._species = stream._species
        self._species_dict  = stream._species_dict
        self._CAS = stream._CAS
        self._IDs = stream._IDs
        self._MW = stream._MW

    # Forward pipping
    def __sub__(self, index):
        if isinstance(index, int):
            return Sink(self, index)
        elif isinstance(index, Stream):
            raise TypeError(f"Unsupported operand type(s) for -: '{type(self)}' and '{type(index)}'")
        return index.__rsub__(self)
        
    def __rsub__(self, index):
        if isinstance(index, int):
            return Source(self, index)
        elif isinstance(index, Stream):
            raise TypeError(f"Unsupported operand type(s) for -: '{type(self)}' and '{type(index)}'")
        return index.__sub__(self)

    # Backward pipping    
    __pow__ = __sub__
    __rpow__ = __rsub__
    
    @property
    def species(self):
        """[Species] Contains pure component thermodynamic properties for computing overall properties of Stream instances."""
        return self._species
    
    @property
    def ID(self):
        """Unique identification (str). If set as '', it will choose a default ID."""
        return self._ID

    @ID.setter
    def ID(self, ID):
        stream = find.stream
        ID_old = self._ID
        if ID_old and ID_old in stream: del stream[ID_old]
        if ID == '':
            # Select a default ID if requested
            letter, number = self._default_ID
            self._default_ID[1] += 1
            num = str(number)
            ID = letter + num
            self._ID = ID
            stream[ID] = self
        elif ID:
            ID = ID.replace(' ', '_')
            ID_words = ID.split('_')
            if not all(word.isalnum() for word in ID_words):
                raise ValueError('ID cannot have any special characters')
            stream[ID] = self
            self._ID = ID

    @property
    def cost(self):
        """Cost of stream (USD/hr)"""
        return self.price*self.massnet

    @property
    def MW(self):
        """Molecular weight of all species (array g/mol):     

        >>> s2.MW 
        tuple_array([46.06844, 18.01528])
        """
        return self._MW.view(tuple_array)

    ### Flow properties ###
    
    def indices(self, *species, CAS=False):
        """Return flow indices of specified species.

        **Parameters**

             ***species:** [iterable] species IDs or CAS numbers.
             
             **CAS:** [bool] True if species are CAS numbers.

        Example

        Indices by ID:
        
        >>> s1.indices('Water', 'Ethanol')
        [1, 0]

        Indices by CAS number:
        
        >>> s1.indices('7732-18-5', '64-17-5', CAS=True):
        [1, 0]

        """
        try:
            lookup = self._CAS.index if CAS else self._IDs.index
            return [lookup(i) for i in species]
        except ValueError as VE:
            if not self._species:
                raise RuntimeError(f'No species defined in stream.')
            if all(isinstance(i, str) for i in species):
                if CAS:
                    stream_species = self._CAS
                    start = 'CAS numbers '
                else:
                    stream_species = self._IDs
                    start = 'IDs '
            else:
                raise TypeError(f'species must all be str objects')
            not_included = tuple(i for i in species if i not in stream_species)
            if not_included: raise ValueError(f'{start}{not_included} not defined in stream species')
            else: raise VE
                
    # Molar flow
    @property
    def mol(self):
        """Array of molar flow rates (kmol/hr):

        >>> s1.mol
        material_array([0, 2])
        >>> s1.mol = [1, 2]
        >>> s1.mol
        material_array([1, 2])
        """
        return self._mol

    @mol.setter
    def mol(self, val):
        try: self._mol[:] = val
        except ValueError as value_error:
            if len(val) != self._Nspecies:
                raise ValueError(f'Length of flow ({len(val)} given) must be the same as the number of species ({self._Nspecies})')
            else:
                raise value_error

    @property
    def molfrac(self):
        """Array of molar fractions.

        >>> s1.molfrac
        tuple_array([0., 1.])
        """
        return fraction(self._molarray).view(tuple_array)

    @property
    def molnet(self):
        """Net molar flow rate (kmol/hr)

        >>> s2.molnet
        3
        """
        return self._molarray.sum()

    # Mass flow
    @property
    def mass(self):
        """Array of mass flow rates (kg/hr)

        >>> s1.mass
        property_array([ 0.   , 36.031])
        """
        return self._mass

    @mass.setter
    def mass(self, val):
        self._mass[:] = val

    @property
    def massfrac(self):
        """Array of mass fractions.

        >>> s1.massfrac
        tuple_array([0, 1])
        """
        return fraction(self._molarray * self._MW).view(tuple_array)

    @property
    def massnet(self):
        """Net mass flow rate (kg/hr)

        >>> s2.massnet
        82.099
        """
        return (self._MW * self._molarray).sum()

    # Volumetric flow
    @property
    def vol(self):
        """Array of volumetric flow rates as a function of T and P (m^3/hr).

        >>> s2.vol
        property_array([0.059, 0.036])
        
        """
        return self._vol

    @vol.setter
    def vol(self, val):
        self._vol[:] = val

    @property
    def volfrac(self):
        """Array of volumetric fractions as a function of T and P.

        >>> s2.volfrac
        tuple_array([0.619, 0.381])
        """
        return fraction(self._vol).view(tuple_array)
        
    @property
    def volnet(self):
        """Net volumetric flow rate as a function of T and P (m^3/hr).

        >>> s2.volnet # a liquid stream
        0.09475552896916632
        >>> s3.volnet # a gas stream
        97.32699378482434
        """
        return self._prop_molar_flownet('Vm')*1000

    ### Energy flows ###

    # Enthalpy
    @property
    def H(self):
        """Enthalpy flow rate as a function of T, excluding formation energies (kJ/hr).

        >>> s1.H # The stream is at the reference state
        0.0
        >>> s1.H = 1000
        >>> s1.T
        304.80600901692236
        >>> s1.H
        1002.1604769775776

        .. Note:: The solver reaches a level of tolerance and the new temperature is an approximation.
        """
        return self._prop_molar_flownet('H')

    @H.setter
    def H(self, H):
        if (self._molarray == 0).all():
            if H != 0: raise ValueError(f"Cannot set nonzero enthalpy to empty stream.")
            else: return
        
        if self.lazy_energy_balance:
            self.T += (H - self.H)/self.C    
        else:
            # First approximation
            T_old = self.T
            C = self.C
            self.T += (H - self.H)/C
        
            # Solve enthalpy by iteration
            it = 1
            while abs(self.T - T_old) > 0.01:
                T_old = self.T
                self.T += (H - self.H)/C
                it += 1
                if it > 40:
                    raise SolverError(f"Could not solve temperature given enthalpy")

    @property
    def Hf(self):
        """Heat of formation flow rate (kJ/hr).

        >>> s1.Hf
        -483640.0
        """
        return self._prop_molar_flownet('Hfm')

    @property
    def Hc(self):
        """Heat of combustion flow rate (kJ/hr).

        >>> s2.Hc
        -1409983.0
        """
        return self._prop_molar_flownet('Hc')
        
    # Entropy
    @property
    def S(self):
        """Entropy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> s1.S # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.S
        10.642980522582695
        """
        return self._prop_molar_flownet('S')

    # Gibbs free energy
    @property
    def G(self):
        """Gibbs free energy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> s1.G # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.G
        -117.6450204179555
        """
        return self.H - self.S*self.T

    # Internal energy
    @property
    def U(self):
        """Internal energy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> s1.U # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.U
        -409.16014979602505
        """
        return self.H - self.P*self.volnet

    # Helmholtz
    @property
    def A(self):
        """Helmholtz energy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> s1.A # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.A
        -3814.9139170224876
        """
        return self.U - self.T*self.S

    # Capacity flow rate
    @property
    def C(self):
        """Heat capacity flow rate as a function of T (kJ/K/hr).

        >>> s2.C
        262.74816631655267
        """
        return self._prop_molar_flownet('Cpm')

    # Material properties
    @property
    def Cp(self):
        """Specific heat capacity as a function of T (kJ/kg/K).

        >>> s1.Cp
        4.180597021827335
        """
        return self.Cpm*self.molnet/self.massnet

    @property
    def Cpm(self):
        """Molar heat capacity as a function of T (kJ/kmol/K).

        >>> s1.Cpm # (J/mol/K)
        75.31462591538556
        """
        return self.C/self.molnet

    @property
    def Vm(self):
        """Molar volume as a function of T and P (m^3/mol).

        >>> s1.Vm
        1.8069039870122814e-05
        """
        return self._prop_molar_mean('Vm')

    @property
    def rho(self):
        """Density as a function of T (kg/m^3).

        >>> s1.rho
        997.0247522552814
        """
        return self.massnet/self.volnet

    @property
    def rhom(self):
        """Molar density as a function of T and P (mol/m^3).

        >>> s1.rhom
        55343.283715561534
        """
        return self.molnet/self._prop_molar_flownet('Vm')

    @property
    def nu(self):
        """Kinematic viscosity as a function of T and P (m^2/s).

        >>> s1.nu
        9.154438858391748e-07
        """
        return self.mu/self.rho

    @property
    def mu(self):
        """Hydraulic viscosity as a function of T and P (Pa*s).

        >>> s1.mu
        0.0009127202134824155
        """
        # Katti, P.K.; Chaudhri, M.M. (1964). "Viscosities of Binary Mixtures of Benzyl Acetate with Dioxane, Aniline, and m-Cresol". Journal of Chemical and Engineering Data. 9 (1964): 442–443.
        molfrac = self.molfrac
        mus = np.array(self._prop_list('mu'))
        Vms = np.array(self._prop_list('Vm'))
        Vm = self.Vm
        pos = np.where(molfrac != 0)
        return exp((molfrac[pos]*ln(mus[pos]*Vms[pos])).sum())/Vm 

    @property
    def k(self):
        """Thermal conductivity as a function of T and P (W/m/k).

        >>> s1.k
        0.5942044328004411
        """
        return self._prop_molar_mean('k')

    @property
    def alpha(self):
        """Thermal diffusivity as a function of T and P (m^2/s).

        >>> s1.alpha
        1.4255801521655763e-07
        """
        return self._prop_molar_mean('alpha')

    @property
    def sigma(self):
        """Surface tension as a function of T (N/m).

        >>> s1.sigma
        0.07205503890847455
        """
        return self._prop_molar_mean('sigma')

    @property
    def Pr(self):
        """Prandtl number as a function of T and P (non-dimensional).

        >>> s1.Pr
        6.421553249380879
        """
        return self._prop_molar_mean('Pr')

    # Other properties
    @property
    def source(self):
        """Unit source."""
        return self._source

    @property
    def sink(self):
        """Unit sink."""
        return self._sink

    @property
    def nonzero_species(self):
        """Flow indices and species that have a non-zero flow rate.

        >>> s1.nonzero_species
        [1], ['Water']
        """
        return nonzero_species(self._num_IDs, self.mol)
    
    def quantity(self, prop_ID):
        """Return a property as a Quantity object as described in the `pint package <https://pint.readthedocs.io/en/latest/>`__ 

        **Parameters**

             **prop_ID:** [str] name of the property (e.g. 'mol', 'H', 'k' ...)

        Example:

        >>> s1.quantity('Cp')
        <Quantity(4.180597021827335, 'joule / gram / kelvin')>

        """
        attr = getattr(self, prop_ID)
        if isinstance(attr, np.ndarray):
            attr = np.array(attr.astype(float))
        return _Q(attr, self.units[prop_ID])

    # Derivative of a property and sensivity analysis
    def derivative(self, prop_ID, var_ID, index=None):
        """Return the derivative of property 'prop_ID' and variable 'var_ID'. If the property given by var_ID is an array, use index to specify which element.

         **Parameters**

             **prop_ID:** [str] Name of the property (e.g. 'rho', 'Cp', 'H', 'volnet' ...)

             **var_ID:** [str] Name of the variable (e.g 'T', 'P', 'molnet' ...)

             **index:** [array_like or None] Optional indices if the variable is an array

        Example

        >>> s1.derivative('rho', 'T')
        -0.4010588564718449

        """
        if index is not None:
            def getvar():
                return getattr(self, var_ID)[index]

            def setvar(xf):
                getattr(self, var_ID)[index] = xf
        else:
            def getvar():
                return getattr(self, var_ID)

            def setvar(xf):
                setattr(self, var_ID, xf)
        x0 = getvar()
        y0 = getattr(self, prop_ID)
        xf = x0 + 10**-6
        setvar(xf)
        yf = getattr(self, prop_ID)
        setvar(x0)
        return (yf-y0)/(xf-x0)

    # General methods for getting ideal mixture properties
    def _prop_list(self, prop_ID):
        """Return component property list."""
        out = np.zeros(self._Nspecies)
        getattr_ = getattr
        P = self.P
        T = self.T
        phase = self.phase
        ic = self._num_compounds.__iter__().__next__
        for m in self._molarray:
            if m: 
                i, c = ic()
                c.P = P
                c.T = T
                c.phase = phase
                prop = getattr_(c, prop_ID)
                out[i] = prop if prop else 0
            else: ic()
        return out

    def _prop_molar_flow(self, prop_ID):
        """Return array of component properties * kmol/hr."""
        return self._prop_list(prop_ID) * self._molarray

    def _prop_molar_flownet(self, prop_ID):
        """Return sum of component properties * kmol/hr."""
        return self._prop_molar_flow(prop_ID).sum()

    def _prop_molar_mean(self, prop_ID):
        """Return molar weighted average property."""
        return self._prop_molar_flownet(prop_ID) / self.molnet

    # Specifications
    @staticmethod
    def like(stream, ID=''):
        """Create either a Stream or MixedStream object just like the given stream, depending on whether multiple phases are present."""
        s = stream
        phase = s.phase
        if len(phase) == 1:
            out = Stream(ID, flow=s.mol, species=s._species,
                         phase=phase, T=s.T, P=s.P)
        else:
            out = MS.MixedStream(ID,
                                 solid_flow=s._solid_mol,
                                 liquid_flow=s._liquid_mol,
                                 LIQUID_flow=s._LIQUID_mol,
                                 vapor_flow=s._vapor_mol,
                                 species=stream._species,
                                 T=s.T, P=s.P)
        return out
    
    def copylike(self, stream):
        """Copy mol, T, P, and phase of stream to self.

        >>> s1.copylike(s3)
        >>> s1.show()
        Stream: s1
         phase: 'g', T: 400.00 K, P: 101325 Pa
         flow (kmol/hr): Ethanol  1
                         Water    2
        """
        phase = stream.phase
        if len(phase) == 1:
            if not (self._species is stream._species):
                raise ValueError('species must be the same to copy stream specifications.')
            self._mol[:] = copy.copy(stream.mol)
            self.P = stream.P
            self.phase = phase
            self.T = stream.T
        else:
            self.enable_phases()
            self.copylike(stream)

    def empty(self):
        """Set flow rates to zero

        >>> s1.empty()
        >>> s1.mol
        material_array([0, 0])
        """
        self._molarray[:] = 0
        
    def _equilibrium_species(self):
        """Return species and indexes of species in equilibrium."""
        species = []; index = []; ic = self._num_compounds.__iter__().__next__
        for m in self.mol:
            if m:
                i, c = ic()
                if c.UNIFAC_Dortmund_groups and c.Tb not in (inf, None):
                    species.append(c); index.append(i)
            else: ic()
        return species, index

    def _heavy_species(self):
        """Return species and indexes of heavy species not in equilibrium."""
        species = []; index = []; ic = self._num_compounds.__iter__().__next__
        for m in self.mol:
            if m:
                i, c = ic()
                if c.Tb in (inf, None) or not c.UNIFAC_Dortmund_groups:
                    species.append(c); index.append(i)
            else: ic()
        return species, index

    def _light_species(self):
        """Return species and indexes of light species not in equilibrium."""
        species = []; index = []; ic = self._num_compounds.__iter__().__next__
        for m in self.mol:
            if m:
                i, c = ic()
                if c.Tb is -inf:
                    species.append(c); index.append(i)
            else: ic()
        return species, index

    # Equilibrium
    def _equilibrium_args(self):
        """Return default arguments for equilibrium calculations."""        
        # Choose species that have UNIFAC groups
        species, index = self._equilibrium_species()
        
        # Get the molar fraction for those species that are not zero
        mol = self._molarray[index]
        z = mol/mol.sum()

        return tuple(species), z, self.P
    
    def bubble_point(self):
        """Bubble point at current composition and Pressure.

        **Returns**

            **T_bubble:** [float] The bubble point temperature

            **y:** [numpy ndarray] The composition of the vapor phase considering only the equilibrium species.
             
            **eq_species:** tuple[str] Species in equilibrium.

        >>> stream = Stream(flow=(0.6, 0.4),
        ...                 species=Species('Ethanol', 'Water'))
        >>> stream.bubble_point()
        (352.2820850833474, array([0.703, 0.297]), (<Chemical: Ethanol>, <Chemical: Water>))

        """
        # If just one specie in equilibrium, return Tsat
        species, x, P = self._equilibrium_args()
        N_species = len(species)
        if N_species == 1:
            return species[0].Tsat(P), 1, species
        elif N_species == 0:
            raise EquilibriumError('No species available for phase equilibrium')
        
        # Solve and return bubble point
        x, P = self._bubble_point(species, x, P)
        return x, P, species
    
    def _bubble_point_error(self, T, species, x, P):
        """Bubble point given x and P"""
        gamma = self.activity_coefficients(species, x, T)
        Psat = [s.VaporPressure(T) for s in species]
        self._yP =  x * Psat * gamma
        return 1 - self._yP.sum()/P
    
    def _bubble_point(self, species, x, P):
        """Bubble point at given composition and Pressure

        **Parameters**

            **species:** tuple[Compound] Species corresponding to x.

            **x:** [array_like] The composition of the liquid phase.

            **P:** [float] The pressure in Pascals.
        
        **Returns**

            **T_bubble:** [float] The bubble point temperature

            **y:** [numpy ndarray] The composition of the vapor phase considering only the equilibrium species.
             
            **eq_species:** tuple[str] Species in equilibrium.

        >>> s1._bubble_point(species=Species('Ethanol', 'Water'),
        ...                 x=(0.6, 0.4), P=101325)
        (352.2820850833474, array([0.703, 0.297]))
        
        """
        # Setup functions to calculate vapor pressure and activity coefficients
        x = np.array(x)
        x /= x.sum()
        
        # Retrive cached info
        # Even if different composition, use previous bubble point as guess
        cached = self._bubble_cached.get(species) 
        if cached:
            cP, T_bubble, y_bubble, cx = cached # c means cached
            if abs(x - cx).sum() < 1e-6 and abs(cP - P) < 1:
                # Return cached data
                return T_bubble, y_bubble
        else:
            # Initial temperature guess
            T_bubble = (x * [s.Tb for s in species]).sum()

        # Solve and return
        T_bubble = newton(self._bubble_point_error, T_bubble,
                          args=(species, x, P), tol= 1e-7)
        y_bubble = self._yP / P
        y_bubble /= y_bubble.sum()
        self._bubble_cached[species] = (P, T_bubble, y_bubble, x)
        return T_bubble, y_bubble

    def dew_point(self):
        """Dew point at current composition and Pressure.
        
        **Returns**

            **T_dew:** [float] The dew point temperature

            **x:** [numpy array] The composition of the liquid phase considering only the equilibrium species.
             
            **eq_species:** tuple[str] Species in equilibrium.

        >>> stream = Stream(flow=(0.5, 0.5),
        ...                 species=Species('Ethanol', 'Water'))
        >>> stream.dew_point()
        (353.7420742149825, array([0.556, 0.444]), (<Chemical: Ethanol>, <Chemical: Water>))
        
        """
        # If just one specie in equilibrium, return Tsat
        species, y, P = self._equilibrium_args()
        N_species = len(species)
        if N_species == 1:
            return species[0].Tsat(P), 1, species
        elif N_species == 0:
            raise EquilibriumError('No species available for phase equilibrium')
        
        # Solve and return dew point
        y, P = self._dew_point(species, y, P)
        return y, P, species
    
    def _dew_point_error_T(self, T, x, y, P, species):
            """Error for dew point temperature, constant x and P."""
            Psat = np.array([s.VaporPressure(T) for s in species])
            gamma = self.activity_coefficients(species, x, T)
            self._y_over_Psat_gamma = y/(Psat*gamma)
            return 1 - self._y_over_Psat_gamma.sum() * P
    
    def _dew_point_error_x(self, x, y, P, species):
        """Error function for dew point composition with constant y and P."""
        x_dew = x/x.sum()
        self._T_dew = newton(self._dew_point_error_T, self._T_dew,
                             args=(x, y, P, species), tol= 1e-7)
        return abs(x_dew - self._y_over_Psat_gamma*P)
    
    def _dew_point(self, species, y, P):
        """Dew point given y and P.

        **Parameters**

            **species:** tuple[Compound] Species corresponding to x.

            **y:** [array_like] The composition of the vapor phase.

            **P:** [float] The pressure in Pascals.

        **Returns**

            **T_dew:** [float] The dew point temperature

            **x:** [numpy array] The composition of the liquid phase considering only the equilibrium species.
             
            **eq_species:** tuple[str] Species in equilibrium.


        >>> s1.dew_point(species=Species('Ethanol', 'Water'),
        ...              y=(0.5, 0.5), P=101325)
        (353.7420742149825, array([0.556, 0.444]))
        """
        y = np.array(y)
        y /= y.sum()
        
        # Retrive cached info
        # Even if different composition, use previous bubble point as guess
        cached = self._dew_cached.get(species)
        if cached:
            cP, self._T_dew, cy, x_dew = cached # c means cached
            if abs(y - cy).sum() < 1e-6 and abs(cP - P) < 1:
                # Return cached data
                return self._T_dew, x_dew 
        else:
            # Initial temperature guess
            self._T_dew = (y * [s.Tb for s in species]).sum()
            x_dew = y

        # Get min and max splits
        Nspecies = len(species)
        min_ = np.zeros(Nspecies)
        max_ = np.ones(Nspecies)

        # Find x by least squares
        try: x_dew = least_squares(self._dew_point_error_x, x_dew,
                                   bounds=(min_, max_),
                                   args=(y, P, species),
                                   xtol= 1e-6).x
        except: pass
        self._dew_cached[species] = (P, self._T_dew, y, x_dew)
        return self._T_dew, x_dew

    # Dimensionless number methods
    def Re(self, L, A=None):
        """Return Reynolds number.

        **Parameters**

             **L:** [float] Characteristic lenght (m)

             **A:** [float] Cross-sectional area (m^2). If A is None, assume flow through a cylindrical pipe.

        Example

        >>> s1.Re(0.2, 0.031416)
        373.43769007101764
        """
        volnet = _Q(self.volnet, self.units['vol']).to('m^3/s').magnitude
        if A is None:  # Assume pipe
            return 4*volnet/(self.nu*np.pi*L)
        return volnet*L/(self.nu*A)

    # Class methods
    @classmethod
    def _helpdata(cls, prop):
        """Return information related to a property.

        **Parameters**

             **prop:** [str] name or description of a property

        Return

             **out:** tupple[str] where elements are:
                  * **ID:** [str] name of the property

                  * **Description:** [str] description of the property

                  * **Dependency:** [str] TP dependency

                  * **Units:** [str] units of measure

                  * **Datatype:** [str] type returned by property

        Example

        >>> Stream._helplist('k')
        ('k', 'thermal conductivity', 'TP', 'W/m/K', 'float')

        >>> Stream._helplist('viscosity')
        (('nu', 'kinematic viscosity', 'TP', 'm^2/s', 'float'),
         ('mu', 'hydraulic viscosity', 'TP', 'Pa*s', 'float'))

        >>> Stream._helplist('kmol/hr')
        (('mol', 'molar flow rates', '', 'kmol/hr', 'np.array'),
         ('molnet', 'net molar flow rate', '', 'kmol/hr', 'float'))
        """
        # Compare by both ID and description
        for index in (0, 1):
            # Search for exact matches
            for l in cls._prop_info:
                if prop == l[index]: return l
        
        out = []
        for index in (0, 1):    
            # If no matches found, search harder
            for l in cls._prop_info:
                if prop.lower() in l[index].lower(): out.append(l)
        return out

    @classmethod
    def help(cls, prop):
        """Print information related to a property.

        **Parameters**

             **prop:** [str] name or description of a property     

        Example

        >>> Stream.help('rho')
        rho: [float] Density as a function of T and P (kg/m^3).
        >>> Stream.help('density')
        rho: [float] Density as a function of T and P (kg/m^3).
        """
        data = cls._helpdata(prop)
        if data: _print_helpdata(data)
        else: print(f"No matching property '{prop}'.")

    @staticmethod
    def eqT(streams, T_guess=None, Q_in=0, approximate=True):
        """Bring all streams to temperature equilibrium.

        **Parameters**

            **streams:** [Stream] All streams
        
            **T_guess:** [float] Equilibrium temperature guess (K)

            **Q_in:** [float] Heat addition to streams (kJ/hr)     
            
            **approximate:** [bool] If True, approximate energy balance with constant heat capacity

        **Return**

             **T:** [float] New temperature (K)

        >>> Stream.eqT((s1, s2, s3))
        >>> (s1.T, s2.T, s3.T)
        (346.06735249999997,  346.06735249999997,  346.06735249999997)
        """
        sum_ = sum
        # Set up problem
        C = sum_(s.C for s in streams)
        H_ = sum_(s.H for s in streams) + Q_in
        
        # Set initial temperature
        if not T_guess:
            T_guess = sum_(s.T for s in streams)/len(streams)
        for s in streams:
            s.T = T_guess
        
        # Find new T
        H = sum_(s.H for s in streams)
        T = T_guess + (H_ - H)/C
        
        # Update
        T_guess = T
        for s in streams:
            s.T = T_guess
        if approximate: return
        
        # Solve enthalpy by iteration
        it = 1
        while abs(T - T_guess) > 0.01:
            # Calculate
            H_ = sum_(s.H for s in streams)
            T = T_guess + (H_- H)/C
            
            # Check iteration
            if it > 40:
                raise SolverError(f"Could not solve temperature.")
            
            # Update
            T_guess = T
            for s in streams:
                s.T = T_guess
            it += 1

    @staticmethod
    def sum(s_sum, streams):
        """Mixes streams and sets resulting mass and energy in 's_sum'. Assumes the same pressure as streams[0], and no phase equilibrium. If 's_sum' is a Stream object, it assumes the phase does not change.

        **Parameters** 

             **s_sum:** [Stream] Container for the resulting mixture of streams

             **streams:** [tuple] Stream objects to be mixed
             
        **Return**

             **s_sum:** same object as in arguments

        .. Note:: Ignores contents of s_sum as if it was empty

        Example

        >>> s2.T = 340 # Set higher temperature to show energy balance
        >>> s_sum = Stream.sum(Stream('s_sum'), s1, s2)
        >>> s_sum.show()
        Stream: s_sum
         phase: 'l', T: 325.32 K, P: 101325 Pa
         flow (kmol/hr): Ethanol  1
                         Water    4
        """
        # Check if energy balance is required
        T_init = streams[0].T
        Ts = np.array([s.T for s in streams[1:]])
        energy_balance = any(T_init != Ts)
        if energy_balance: H = sum(s.H for s in streams)
        else: s_sum.T = T_init

        # Copy starting values
        s_sum.copylike(streams[0])
        streams = streams[1:]

        # Mass balance
        inst = isinstance
        if inst(s_sum, MS.MixedStream):
            # For MixedStream objects
            for s in streams:
                if inst(s, MS.MixedStream):
                    s_sum._molarray += s._molarray
                elif inst(s, Stream):
                    s_sum._phases_molflow[s.phase] += s._molarray
        elif inst(s_sum, Stream):
            # For Stream objects
            for s in streams:
                s_sum._molarray[:] += s.mol
        else:
            raise TypeError('Must pass a valid Stream object')

        # Energy Balance
        if energy_balance: s_sum.H = H
        return s_sum

    # MixedStream compatibility
    def enable_phases(self):
        """Cast stream into a MixedStream object."""
        default_phase = self.phase
        self.__class__ = MS.MixedStream
        self._setflows(np.zeros((4, self._Nspecies)))
        self._default_phase = default_phase
        self._phases_molflow[default_phase][:] = self._mol

    def disable_phases(self, phase):
        """Cast stream into a Stream object.
        
        **Parameters**
        
            **phase:** {'s', 'l', 'g'} desired phase of stream
            
        """
        self.phase = phase
            
    def VLE(self, species_IDs=None, LNK=None, HNK=None, P=None,
            Qin=None, T=None, V=None, x=None, y=None):
        self.enable_phases()
        self.VLE(species_IDs, LNK, HNK, P, Qin, T, V, x, y)
        
    def LLE(self, species_IDs=None, split=None, lNK=(), LNK=(),
            solvents=(), solvent_split=(),
            P=None, T=None, Qin=None):
        self.enable_phases()
        self.LLE(species_IDs, split, lNK, LNK,
                 solvents, solvent_split, P, T, Qin)

    def _info_header(self):
        """Return stream information header."""
        # First line
        unit = self._source
        if unit is None:
            source = ''
        else:
            source = f'  from  {type(unit).__name__}-{unit}'
        unit = self._sink
        if unit is None:
            sink = ''
        else:
            sink = f'  to  {type(unit).__name__}-{unit}'
        if self.ID:
            return f"{type(self).__name__}: {self.ID}{source}{sink}"
        else:
            return f"{type(self).__name__}{source}{sink}"

    def _info_units(self, show_units):
        """Return unit format selection."""
        return (show_units.get('T', 'K'),
                show_units.get('P', 'Pa'),
                show_units.get('flow', 'kmol/hr'),
                show_units.get('fraction', False))
        
    def _info_phaseTP(self, phases, T_units, P_units):
        T = _Q(self.T, self.units['T']).to(T_units).magnitude
        P = _Q(self.P, self.units['P']).to(P_units).magnitude
        return f" phase: '{phases}', T: {T:.5g} {T_units}, P: {P:.6g} {P_units}\n"

    # Representation
    def _info(self, **show_units):
        """Return string with all specifications."""
        units = self.units
        basic_info = self._info_header() + '\n'
        T_units, P_units, flow_units, fraction = self._info_units(show_units)
        basic_info += self._info_phaseTP(self.phase, T_units, P_units)
        
        # Start of third line (flow rates)
        flow_dim = _Q(0, flow_units).dimensionality
        if fraction:
            if flow_dim == mol_flow_dim:
                flownet = _Q(self.molnet, units['molnet']).to(flow_units).magnitude
                flow = 'molfrac'
            elif flow_dim == mass_flow_dim:
                flownet = _Q(self.massnet, units['massnet']).to(flow_units).magnitude
                flow = 'massfrac'
            elif flow_dim == vol_flow_dim:
                flownet = _Q(self.volnet, units['volnet']).to(flow_units).magnitude
                flow = 'volfrac'
            else:
                raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{flow_dim}'.")
            beginning = ' flow: '
            end = f'\n*{flownet:.3g} {flow_units} '
        else:
            beginning = f' flow ({flow_units}): '
            end = ''
            if flow_dim == mol_flow_dim:
                flow = 'mol'
            elif flow_dim == mass_flow_dim:
                flow = 'mass'
            elif flow_dim == vol_flow_dim:
                flow = 'vol'
            else:
                raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{flow_dim}'.")
        # Remaining lines (all flow rates)
        new_line_spaces = len(beginning) * ' '
        nonzero, species = self.nonzero_species
        if 'frac' in flow:
            flow = getattr(self, flow)[nonzero]
        else:
            flow = _Q(getattr(self, flow)[nonzero], units[flow]).to(flow_units).magnitude
        len_ = len(nonzero)
        if len_ == 0:
            return basic_info + ' flow: 0'
        else:
            flowrates = ''
            lengths = [len(sp) for sp in species]
            maxlen = max(lengths) + 1
            for i in range(len_-1):
                spaces = ' ' * (maxlen - lengths[i])
                flowrates += species[i] + spaces + \
                    f' {flow[i]:.3g}\n' + new_line_spaces
            spaces = ' ' * (maxlen - lengths[len_-1])
            flowrates += species[len_-1] + spaces + f' {flow[len_-1]:.3g}'
        
        return basic_info + beginning + flowrates + end.replace('*', new_line_spaces + 'net' + (maxlen-4)*' ' + '  ')

    def show(self, **show_units):
        """Print all specifications."""
        print(self._info(**show_units))

    def __str__(self):
        if self.ID: return self.ID
        else: return type(self).__name__

    def __repr__(self):
        if self.ID: return f'<{type(self).__name__}: {self.ID}>'
        else: return f'<{type(self).__name__}>'
        
        
from . import _mixed_stream as MS



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:15:20 2018

@author: yoelr
"""
from .. import Unit
from ._splitter import Splitter
from .designtools import vacuum_system
import numpy as np

class RotaryVacuumFilter(Unit):
    _has_power_utility = True
    
    #: Revolutions per second
    rps = 20/3600
    
    #: Radius of the vessel (m)
    radius = 1
    
    #: For crystals (lb/day-ft^2)
    filter_rate = 6000
    _kwargs = {'split': None,
               'P_suction': 100} # Pa
    _bounds = {'Individual area': (10, 800)}
    _units = {'Area': 'ft^2',
              'Individual area': 'ft^2'}
    
    #: Efficiency of the vacuum pump
    power_efficiency = 0.9
    
    @staticmethod
    def _calc_Area(flow:'kg/hr', filter_rate:'lb/day-ft^2') -> 'ft^2':
        flow *= 52.91
        return flow/filter_rate
        
    _run = Splitter._run
    
    def _design(self):
        flow = sum(stream.massnet for stream in self.outs)
        Area = self._calc_Area(flow, self.filter_rate)
        self._results['Design']['Area'] = Area
        
    def _cost(self):
        results = self._results
        Design = results['Design']
        Area = Design['Area']
        ub = self._bounds['Individual area'][1]
        N_vessels = np.ceil(Area/ub)
        self._power(Area, N_vessels)
        iArea = Area/N_vessels # individual vessel
        Design['Individual area'] = iArea
        
        Cost = np.exp(11.796-0.1905*np.log(iArea)+0.0554*(np.log(iArea))**2)
        results['Cost']['Cost of vessels'] = N_vessels*Cost*self.CEPCI/567
    
    def _power(self, area, N_vessels) :
        r = self._results
        s_cake, s_vacuumed = self.outs
        P_suction = self._kwargs['P_suction'] # Max allowable pressure drop
        
        # Weight of empty plate
        mass_plates = 10*N_vessels
        
        # Revolutions per s
        rps = self.rps
        
        # Cake volumetric flow meter per sec
        Volcake = s_cake.volnet / 3600
        
        # Thickness of cake layer, assumed uniform and constant
        thCake = Volcake / rps;
        
        # Mass of Cake
        mass_cake = thCake * s_cake.rho
        radius = self.radius
        cent_a = (2*np.pi*rps)**2*radius
        cent_F = (mass_cake + mass_plates)*cent_a
        work_rot = rps*2*np.pi*radius*cent_F
        Area = r['Design']['Area']
        vol = radius*Area*0.0929/2 # m3
        
        # Assume same volume of air comes in as volume of liquid
        volflow = s_vacuumed.volnet
        massflow = volflow*1.2041 # multiply by density of air kg/m3 
        work_vacuum, r['Cost']['Liquid-ring pump'] = vacuum_system(massflow,
                                                                   volflow,
                                                                   P_suction, vol)
        power = work_rot/self.power_efficiency/1000 + work_vacuum # kW
        self._power_utility(power)
        
        
RVF = RotaryVacuumFilter





# -*- coding: utf-8 -*-
"""
Created on Sat Nov 10 19:27:01 2018

@author: yoelr
"""
from ._compound import Compound
import numpy as np

__all__ = ('DissolvedCompound',)


# %%

class DissolvedCompound(Compound):
    """
    Creates a DissolvedCompound object with the same material properties as "obj" argument for simplified modeling purposes. The phase stays constant as a liquid ('l').

    **Parameters**
    
        **ID:** [str] ID of specie
        
        **CAS:** [str] CAS identification
        
        **obj** [Specie] Should contain analog properties for specie at STP. Defaults to water if None.

        **MW:** Molecular weight of specie
         
        **T:** Temperature (K)
         
        **P:** Pressure (Pa)
         
        **rho:** Density (kg/m^3)
        
        **k:** Thermal conductivity (W/m/K)
        
        **mu:** Hydrolic viscosity (Pa*s)
        
        **sigma:** Surface tension (N/m)
        
        **Hfm:** Heat of formation (J/mol)
        
        **Hc:** Heat of combustion (J/mol)
        
    **Examples**
    
        Create a 'Yeast' compound with the same properties as water:
    
        .. code-block:: python
            
            >>> from biosteam import DissolvedCompound
            >>> Yeast = DissolvedCompound('Yeast', obj=Chemical('Water'))
            >>> Yeast.rho
            997
            
    """
    Tb = np.inf #: No boiling point (K)
    phase_ref = 'l'
    
    def __init__(self, ID, CAS='', obj=None, MW=1, T=298.15, P=101325,
                 Cp=None, rho=None, k=None, mu=None,
                 sigma=None, Pr=None, Hfm=None, Hc=None):
        self.ID = ID
        self.T = T
        self.P = P
        self.MW = MW
        self.Hvapm = 0
        
        # Default values
        Cp_d = 4.18
        rho_d = 997
        k_d = 0.5942
        mu_d = 0.00091272
        sigma_d = 0.072055
        Hfm_d = 0
        Hc_d = 0
        
        # Set object default values
        if obj:
            Cp_d = obj.Cp if obj.Cp else Cp_d
            rho_d = obj.rho if obj.rho else rho_d
            k_d = obj.k if obj.k else k_d
            mu_d = obj.mu if obj.mu else mu_d
            sigma_d = obj.sigma if obj.sigma else sigma_d
            Hfm_d = obj.Hf if obj.Hf else Hfm_d
            Hc_d = obj.Hc if obj.Hc else Hc_d
        
        # Set all values
        self.CAS = CAS if CAS else ID
        self.Cp = Cp if Cp else Cp_d
        self.Cpm = self.Cp * MW
        self.rho = rho if rho else rho_d
        self.rhom = self.rho/MW*1000
        self.Vm = 1/self.rhom
        self.k = k if k else k_d
        self.alpha = self.k/self.rho/self.Cp
        self.mu = mu if mu else mu_d
        self.nu = self.mu/self.rho
        self.sigma = sigma if sigma else sigma_d
        self.Pr = self.nu/self.alpha
        self.UNIFAC_groups = self.UNIFAC_Dortmund_groups = None
        self.Hfm = Hfm if Hfm else Hfm_d
        self.Hc = Hc if Hc else Hc_d
    
    @property
    def Cplm(self):
        return self.Cpm
    @property
    def phase(self):
        """Phase is always 'l' (liquid), regardless."""
        return 'l'
    @phase.setter
    def phase(self, phase): pass


# -*- coding: utf-8 -*-
"""
Created on Wed Mar  6 15:22:37 2019

@author: yoelr
"""
        
from .. import Unit, Stream
from inspect import signature
from .._utils import function

__all__ = ('Block',)

do_nothing = lambda: None
    

# %% Simulation blocks

class Block:
    """
    Create a Block object that can simulate the element and the system downstream. The block can also generate block functions.
    
    **Parameters**
    
        **element:** [Unit or Stream] Element in the system.
        
        **system:** [System] If None, ignore downstream.
        
        **getter:** [function] Getter for block function.
        
        **setter:** [function] Setter for block function.
        
    **Examples**

         :doc:`Block Example`
    
    Create a block object passing a Unit object, and simulate the block:
        
    .. code-block:: python
    
       >>> P1 = Unit('P1')
       >>> block = Block(P1)
       >>> block.show()
       [Block: Unit-P1]
       >>> block.simulate()
    
    .. Note::
    
       Because no system was passed, the block would simulate just the unit.
       
    Create a block object passing both a Unit object and a System object:
        
    .. code-block:: python
    
       >>> P0 = Unit('P0', outs=Stream())
       >>> P1 = Unit('P1', ins=P0-0)
       >>> P2 = Unit('P2', ins=P1-0)
       >>> system = System('', network=(P0, P1, P2))
       >>> block = Block(P1, system)
       >>> block.show()
       [Block: Unit-P1 and downstream]
       >>> block.system.show()
       System: Unit-P1 and downstream
        network: (P1, P2)
    
    .. Note::
    
       The block would simulate Unit-P1 and downstream.
    
    
    Calling a block object with a getter and a setter will return a block function equivalent to:
        
    .. code-block:: python
    
       >>> def blockfunc(args):
       ...     setter(args)
       ...     self.simulate() # self is the Block object
       ...     return getter()  
       
    For example:
        
    .. code-block:: python
    
       >>> # getter and setter functions are hypothetical
       >>> def getter(): pass
       >>> def setter(args): pass
       >>> blockfunc = Block(P1, None)(getter, setter)
       >>> blockfunc
       <function [Block P1] getter(args)>
    
    .. Note::
        
       The function signature matches the setter while the function name matches the getter.
    
    """
    
    _cachedblocks = {}
    __slots__ = ('_system', '_simulate', '_element')
    
    def __new__(cls, element, system=None):
        cached = cls._cachedblocks
        block = cached.get((system, element))
        if block:
            self = block
        else:
            self = object.__new__(cls)
            self.__init__(element, system)
        return self
    
    def __init__(self, element, system=None):
        inst = isinstance
        if inst(element, Stream): unit = element.sink
        elif inst(element, Unit): unit = element
        elif element is None: return self._emptyblock
        else: raise ValueError(f"element must be either a Unit, Stream, or None object, not '{type(element).__name__}'.")
        if system:
            subsys = system._downstream_system(unit)
            simulate = subsys.simulate
        else:
            subsys = system
            if element is Unit: simulate = unit.simulate
            else: simulate = do_nothing
            
        self._system = subsys
        self._simulate = simulate
        self._element = element
        self._cachedblocks[system, element] = self
    
    def _make_blockfunc(self, getter, setter, docfunc):
        # Prevent downstream exceptions
        getter_param = signature(getter).parameters
        if getter_param:
            for p in getter_param.values():
                if '=' not in str(p):
                    raise ValueError(f'{type(self).__name__} getter signature cannot have arguments.')
        sig = signature(setter)
        setter_params = tuple(sig.parameters.keys())
        param = ', '.join(setter_params)
        if len(setter_params) != 1: 
            raise ValueError(f'{type(self).__name__} setter signature must have one and only one argument.')
        
        # Make blockfunc
        name = getter.__name__
        if name[0] == '<': name = 'Lambda'
        str2exec =  (f'def {name}{sig}:    \n'
                   + f'    setter({param}) \n'
                   + f'    simulate()      \n'
                   + f'    return getter()   ')

        globs = {'setter': setter,
                 'simulate': self._simulate,
                 'getter': getter}
        locs = {}
        exec(str2exec, globs, locs)
        blockfunc = locs[name]
        blockfunc.__qualname__ = f'{self} {blockfunc.__qualname__}'
        blockfunc.__doc__ = docfunc.__doc__
        blockfunc._param = param
        blockfunc._block = self
        blockfunc._setter = setter
        blockfunc._getter = getter
        return blockfunc
    
    def __call__(self, getter, setter) -> function:
        return self._make_blockfunc(getter, setter, getter)
    
    @property
    def system(self):
        """System that the block simulates."""
        return self._system
    
    @property
    def element(self):
        """Block element that gets checked for changes before simulation."""
        return self._element
    
    @property
    def simulate(self):
        """Simulate block system if block element has changed. If block element has not changed, simulate only the element."""
        return self._simulate
    
    #: A Block object without a Unit or a System object.
    _emptyblock = None
        
    def _repr(self):
        if self._element:
            return f"{type(self).__name__}: {type(self._element).__name__}-{self._element}"
        else:
            return f"{type(self).__name__}"
    
    def __repr__(self):
        return f'[{self._repr()}]'
    
    def _info(self):
        if self._system:
            return f'[{type(self).__name__}: {self._system}]'
        else:
            return repr(self)

    def show(self):
        print(self._info())

emptyblock = object.__new__(Block)
emptyblock._system = None
emptyblock._element = None
emptyblock._simulate = do_nothing
Block._emptyblock = emptyblock
del emptyblock


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:49:58 2018

@author: yoelr
"""
from ._reactor import Reactor
import numpy as np


class Transesterification(Reactor):
    """Create a transesterification reactor that converts 'Lipid' and 'Methanol' to 'Biodiesel' and 'Glycerol'. Finds the amount of catalyist 'NaOCH3' required and consumes it to 'NaOH' and 'Methanol'.
    
    **Parameters**
    
        **efficiency:** Efficiency of conversion (on a 'Lipid' basis)
        
        **r:** Methanol to lipid molar ratio
        
        **catalyst_molfrac:** Molar fraction of catalyst in methanol feed 
        
        **T:** Operating temperature (K)
    
    **ins**
    
        [0] Lipid feed
        
        [1] Fresh methanol and catalyst
        
    **outs**
    
        [0] Product
    
    **Examples**
    
        :doc:`Transesterification Example`
    
    """
    _units = {'Volume': 'm^3'}
    _kwargs = {'efficiency': None,  # fraction of theoretical conversion
               'r': None,  # Methanol to lipid molar ratio
               'T': None,
               'catalyst_molfrac': None}  # operating temperature (K)
    
    _bounds = {'Volume': (0.1, 20)}
    
    _tau = 1
    _N_ins = 2
    _N_outs = 1
    _N_heat_utilities = 1
    _has_power_utility = True

    #: kW/m3
    electricity_rate = 1.5

    def _run(self):
        feed, fresh_Methanol = self.ins
        out = self.outs[0]
        eff, r, T, catalyst_molfrac = (self._kwargs[i] for i in (
            'efficiency', 'r', 'T', 'catalyst_molfrac'))
        sp_index = feed._IDs.index

        # Reactant positions
        lipid_pos = sp_index('Lipid')
        Methanol_pos = sp_index('Methanol')
        Glycerol_pos = sp_index('Glycerol')
        biodiesel_pos = sp_index('Biodiesel')
        NaOH_pos = sp_index('NaOH')
        NaOCH3_pos = sp_index('NaOCH3')

        lipid = feed.mol[lipid_pos]

        # Reaction conversions by mol (to add to the feed)
        lipid_ch = -eff*lipid  # - free_lipid_ch/3
        bd_ch = eff*lipid*3
        gly_ch = -lipid_ch
        dummy = r*lipid
        NaOCH3_in = dummy*catalyst_molfrac  # from methanol stream
        Methanol_in = dummy - NaOCH3_in  # from methanol stream
        Methanol_ch = dummy + lipid_ch
        NaOCH3_ch = 0
        NaOH_ch = NaOCH3_in

        change_mol = np.array((lipid_ch, Methanol_ch, gly_ch,
                               bd_ch, NaOCH3_ch, NaOH_ch))

        # Output stream
        fresh_Methanol.mol[ [Methanol_pos, NaOCH3_pos] ] = (Methanol_in, NaOCH3_in)
        index = [lipid_pos, Methanol_pos, Glycerol_pos,
                biodiesel_pos, NaOCH3_pos, NaOH_pos]
        out.phase = 'l'
        out.mol[index] = feed.mol[index] + change_mol
        out.T = T
        out.P = feed.P

    def _design(self):
        self._heat_utilities[0](self._Hnet, self.outs[0].T)
        Design = self._results['Design']
        Design['Volume'] = self._tau * self._volnet_out / 0.8
        return Design
        
    def _cost(self):
        results = self._results
        Design = results['Design']
        Cost = results['Cost']
        Volume = Design['Volume']
        Cost['Reactor'] = self.CEPCI/525.4 * 15000 * Volume ** 0.55
        self._power_utility(Volume*self.electricity_rate)
    
        
        



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 11 11:20:42 2018

@author: yoelr
"""
from . import _Q

__all__ = ('PowerUtility',)

class PowerUtility:
    """Create an PowerUtility object that can calculates the cost of power.
    
    **__call__()**
    
       Calculate utility requirements given the essential parameters.
        
        **Parameters**
        
            rate: [float] Power requirement (kW)
            
    **Class Parameters**
    
        **price:** ($/kW-hr)
    
    **Examples**
    
        Create a PowerUtility object:
        
        .. code-block:: python
        
           >>> pu = PowerUtility()
           >>> pu
           <PowerUtility: None>
           
        Call object to calculate cost:
            
        .. code-block:: python
        
           >>> pu(rate=500)
           >>> pu
           <PowerUtility: 500 kW, 30 USD/hr>
           
        Results are accessible:
            
        .. code-block:: python
        
           >>> pu.rate, pu.cost
           (500, 30.)
           
        See the object with different units:
            
        .. code-block:: python
        
           >>> pu.show(rate='BTU/s', cost='USD/yr')
           PowerUtility: rate=474 BTU/s, cost=2.63e+05 USD/yr
    
    """
    _units = dict(rate='kW', cost='USD/hr')
    __slots__ = ('rate', 'cost')
    price = 0.0782 #: USD/kWhr
    
    def __init__(self):
        self.rate = 0
        self.cost = 0
    
    def __call__(self, rate:'kW'):
        """Calculate cost and save. 
        
        **Parameters**
        
            rate: [float] Power requirement (kW)
        
        """
        self.rate = rate
        self.cost = self.price * rate
    
    # Representation
    def _info(self, **show_units):
        # Get units of measure
        su = show_units
        units = self._units
        Rate = su.get('rate') or units['rate']
        Cost = su.get('cost') or units['cost']
        if self.rate:
            rate = _Q(self.rate, units['rate']).to(Rate).magnitude
            cost = _Q(self.cost, units['cost']).to(Cost).magnitude
            return (f'{type(self).__name__}: rate={rate:.3g} {Rate}, cost={cost:.3g} {Cost}')
        else:
            return (f'{type(self).__name__}: None')

    def show(self, **show_units):
        print(self._info(**show_units))

    def __repr__(self):
        units = self._units
        if self.rate:
            Rate = units['rate']
            Cost = units['cost']
            return (f'<{type(self).__name__}: {self.rate:.3g} {Rate}, {self.cost:.3g} {Cost}>')
        else:
            return (f'<{type(self).__name__}: None>')
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:17:05 2018

@author: yoelr
"""

from ._splitter import Splitter
from .. import Unit

__all__ = ['MolecularSieve', 'MolSieve']

class MolecularSieve(Unit):
    """Create an ethanol/water molecular sieve for bioethanol plants. The molecular sieve is modeled as a component wise separator. Costing is based on scaling by the 6/10ths rule from an NREL TEA report [1].
    
    **Parameters**
    
        **split:** *[array_like]* Component wise split in the 0th output stream

    **ins**
    
        [:] Input streams
        
    **outs**
    
        [0] Split stream
        
        [1] Remainder stream
    
    **References**
    
        [1] Process Design and Economics for Biochemical Conversion of Lignocellulosic Biomass to Ethanol Dilute-Acid Pretreatment and Enzymatic Hydrolysis of Corn Stover. D. Humbird, R. Davis, L. Tao, C. Kinchin, D. Hsu, and A. Aden (National Renewable Energy Laboratory Golden, Colorado). P. Schoen, J. Lukas, B. Olthof, M. Worley, D. Sexton, and D. Dudgeon (Harris Group Inc. Seattle, Washington and Atlanta, Georgia)
    
    **Examples**
    
        :doc:`MolecularSieve Example`
    
    """
    _kwargs = Splitter._kwargs
    _run = Splitter._run
    _has_power_utility = True
    
    #: Original Price (USD)
    C_0 = 2601000
    
    #: Original flow rate (kg/hr)
    V_0 = 22687 
    
    #: Scaling exponent
    exp = 0.60 
    
    def _cost(self):
        results = self._results
        Cost = results['Cost']
        CEPCI_old = 521.9 # Original CEPCI (2009)
        CEPCI_new = self.CEPCI
        F_CEPCI = CEPCI_new/CEPCI_old
        
        # Cost molecular sieve vessels with packing
        Cp_old = self.C_0 # Original purchase price (USD)
        flow_old = self.V_0 # Original flow rate (kg/hr)
        flow_new = self._massnet_in
        x = self.exp # Scaling exponent
        S = flow_new/flow_old # Size ratio
        Cp_new = F_CEPCI * Cp_old * S**x
        Cost['Molecular Sieve Cost'] = Cp_new
        
        # Cost pressure filter (pressing compressor)
        Cp_old = 75200
        Cost['Pressure Filter Pressing'] = F_CEPCI * Cp_old * S**x
        power0 = 112*S # kW
        
        # Cost pressure filter (drying compressor)
        N = 2
        Cp_old = 405000
        Cost['Pressure Filter Drying (2)'] = N * F_CEPCI * Cp_old * S**x
        power1 = N*522*S # kW
        
        # Find power utility
        total_power = power0 + power1
        self._power_utility(total_power)
        return Cost
        
MolSieve = MolecularSieve        
        





# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 15:47:26 2018

@author: yoelr
"""

from .. import Unit
from ..exceptions import DesignError
from ._mixer import Mixer
import numpy as np

class Tank(Unit):
    """Abstract Tank class."""
    _units = {'Volume': 'm^3'}
    _N_ins = 1
    _N_outs = 1
    
    #: Abstract residence time attribute
    _tau = None
    _has_proxystream = True

    @property
    def tau(self):
        """Residence time (hr)."""
        return self._tau
    @tau.setter
    def tau(self, tau):
        self._tau = tau

    def _run(self):
            self.outs[0].copylike(self.ins[0])


class StorageTank(Tank):
    """Create a storage tank with volume based on residence time [1].

    .. math::

        V &= \\tau Q \\\\
        C_{fob}^{2007} &= 250000 + 94.2 V (2*10^3 < V < 50*10^3 m^3) \\\\

    **References**

        [1] I.K. Kookos, Preliminary Chemical Process Synthesis and Design, Tziolas Publishing,
Thessalonika, Greece, 2008 (book in Greek).

    """
    #: [float] residence time (hr)
    _tau = 4*7*24

    def _design(self):
        Design = self._results['Design']
        V = self._tau*self._volnet_out
        if V < 50e3:
            Design['N_vessel'] = N = 1
        else:    
            Design['N_vessel'] = N = np.ceil(V/50e3)
        Design['Volume'] = V/N
        return Design

    @staticmethod
    def _calc_vessel_cost(V_vessel: 'm^3', CEPCI):
        if V_vessel < 2e3:
            cost = 65000 + 158.7*V_vessel * CEPCI/525.4
        elif V_vessel < 50e3:
            cost = 250000 + 94.2*V_vessel * CEPCI/525.4
        else:    
            raise DesignError(f"Volume is out of bounds for costing")
        return cost

    def _cost(self):
        Design = self._results['Design']
        V_vessel, N_vessel = Design['Volume'], Design['N_vessel']
        CEPCI = self.CEPCI
        Cost = self._results['Cost']
        Cost['Tank'] = N_vessel * self._calc_vessel_cost(V_vessel/N_vessel, CEPCI)
        return Cost


class MixTank(Tank):
    """Create a mixing tank with volume based on residence time.

    .. math::

       V &= \\frac{\\tau* Q}{0.8}

       C_{fob}^{2007} &= 12080 * V^{0.525} (0.1 < V < 30 m^3)

    References

         [1] I.K. Kookos, Preliminary Chemical Process Synthesis and Design, Tziolas Publishing, Thessalonika, Greece, 2008 (book in Greek).

    """
    _tau = 1
    _N_ins = 2
    _has_power_utility = True
    _has_proxystream = False
    _run = Mixer._run
    _bounds = {'Volume': (0.1, 30)}
    
    #: Electricity rate (kW/m3)
    electricity_rate = 0.591 # About 10 hp/1000gal
    
    def _design(self):
        Design = self._results['Design']
        Design['Volume'] = self._tau * self._volnet_out / 0.8

    def _cost(self):
        V = self._results['Design']['Volume']
        Cost = self._results['Cost']
        Cost['Tank'] = 12080 * V**0.525 * self.CEPCI/525.4
        self._power_utility(self.electricity_rate*V)



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:18:36 2018

@author: yoelr
"""
import numpy as np
from ._splitter import Splitter
from .. import Unit
from ..exceptions import DesignError


class YeastCentrifuge(Unit):
    """Create a yeast centrifuge that separates out solids according to user defined split. Assume a continuous scroll solid bowl. 
    
    **Parameters**
    
        **split:** *[array_like]* Component splits to 0th output stream
        
        **solids:** *tuple[str]* IDs of solids 
    
    **ins**
    
        [:] Input streams
    
    **outs**
    
        [0] Liquid stream
        
        [1] Solids stream
    
    """
    
    _kwargs = {'split': None,
               'solids': None}
    
    _run = Splitter._run
    
    def _cost(self):
        Cost = self._results['Cost']
        solids = self._kwargs['solids']
        index = self.outs[0].indices(*solids)
        mass_solids = 0
        for s in self.ins:
            mass_solids += s.mass[index]
        ts = np.asarray(mass_solids).sum() # Total solids
        ts *= 0.0011023 # To short tons (2000 lbs/hr)
        if 2 < ts < 40:
            Cost['Centrifuge'] = self.CEPCI*68040/567*ts**0.50
        else:
            raise DesignError(f'Solids loading ({ts}) is not within 2 and 40 tonns')
        return Cost



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 14:34:07 2018

@author: yoelr
"""
from .. import Unit, Stream
from .._meta_final import metaFinal

class Splitter(Unit, metaclass=metaFinal):
    """Create a splitter that separates mixed streams based on splits.

    **Parameters**

        **split:** Shoulds be one of the following:
            * *[float]* The fraction of net feed in the 0th output stream
            * *[array_like]* Component wise split of feed to 0th output stream
    
    **ins**
    
        [:] Input streams
        
    **outs**
    
        [0] Split stream
        
        [1] Remainder stream
    
    **Examples**
    
        Create a Splitter object with an ID, any number of input streams, two output streams, and an overall split:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, Splitter
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = Splitter('S1', ins=feed, outs=('out1', 'out2'), split=0.1)
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] out1
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  1
           [1] out2
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  9
          
        Create a Splitter object, but this time with a component wise split:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, Splitter
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = Splitter('S1', ins=feed, outs=('out1', 'out2'),
           ...               split=(0.1, 0.99))
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] out1
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  9.9
           [1] out2
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  0.1
    
    """
    _kwargs = {'split': None}
    results = None
    _has_cost = False
    _N_outs = 2

    def __init__(self, ID='', outs=(), ins=None, split=None):
        self.ID = ID
        self._kwargs = {'split': split}
        self._init_ins(ins)
        self._init_outs(outs)

    def _run(self):
        # Unpack
        split = self._kwargs['split']
        top, bot = self.outs
        if len(self.ins) > 1: Stream.sum(self.outs[0], self.ins)
        else: top.copylike(self.ins[0])
        bot.copylike(top)
        top.mol = top.mol*split
        bot.mol = bot.mol - top.mol
    
    simulate = _run


class InvSplitter(Unit, metaclass=metaFinal):
    """Create a splitter that sets input stream based on output streams. Must have only one input stream. The output streams are the same temperature, pressure and phase as the input.
    """
    line = 'Splitter'
    results = None
    _graphics = Splitter._graphics
    _has_cost = False
    
    def __init__(self, ID='', outs=(), ins=None):
        self.ID = ID
        self._init_ins(ins)
        self._init_outs(outs)
        
    def _run(self):
        feed = self.ins[0]
        feed.mol = self._mol_out
        for out in self.outs:
            out.T, out.P, out.phase = feed.T, feed.P, feed.phase 
            
    simulate = _run



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 21:55:57 2019

@author: yoelr
"""
from ._stream import Stream
from ._mixed_stream import MixedStream

__all__ = ('ProxyStream', 'ProxyMixedStream')

#%% Proxy streams for efficiency

class ProxyStream(Stream):
    """Create a ProxyStream object that shares data with another stream. ProxyStream objects work exactly the same as Stream objects.
    
    **Parameters**
    
        **ID:** [str] ID of proxy stream
        
        **link:** [Stream] proxy stream will share data from this link.
    
    """
    _data = {}
    _link = None
    __slots__ = ()
    
    def __init__(self, ID='', link=None):
        ProxyStream._data[self] = [None, None, None]
        self.ID = ID
        self.link = link

    def copylike(self, stream):
        if not self._link is stream: super().copylike(stream)
    copylike.__doc__ = Stream.copylike.__doc__
    
    @classmethod
    def asproxy(cls, stream):
        """Cast the stream as a proxy object."""
        cls._data[stream] = [stream._source, stream._sink, stream.ID]
        stream.__class__ = cls
        return stream
    
    @property
    def link(self):
        """Stream that the proxy stream is sharing data with."""
        return self._link
    
    @link.setter
    def link(self, stream):
        if not self._link is stream:
            if hasattr(self, 'price'): stream.price = self.price
            self.__dict__ = stream.__dict__
            stream._link = stream
        elif stream and not isinstance(stream, Stream):
            raise TypeError(f"link must be a Stream object or None, not a '{type(stream).__name__}' object.")
        
    @property
    def _ID(self):
        return self._data[self][2]
    
    @_ID.setter
    def _ID(self, ID):
        self._data[self][2] = ID
    
    @property
    def _source(self):
        return self._data[self][0]
    
    @_source.setter
    def _source(self, source):
        self._data[self][0] = source
    
    @property
    def _sink(self):
        return self._data[self][1]
    
    @_sink.setter
    def _sink(self, sink):
        self._data[self][1] = sink
        
    @property
    def P(self):
        if hasattr(self, '_P'): return self._P
        else: return self.__dict__['P']
        
    @P.setter
    def P(self, P):
        self._P = P
    
    def disable_phases(self, phase):
        super().disable_phases(phase)
        self.__class__ = ProxyStream
    disable_phases.__doc__ = Stream.disable_phases.__doc__    
    
    def enable_phases(self):
        super().enable_phases()
        self.__class__ = ProxyMixedStream
    enable_phases.__doc__ = Stream.enable_phases.__doc__

    def _info(self, *args, **kwargs):
        linkinfo = f'\n link: {self.link}'
        if self.__dict__: return super()._info(*args, **kwargs) + linkinfo
        else: return (self._info_header() + linkinfo)
        

class ProxyMixedStream(MixedStream, ProxyStream):
    """Create a ProxyMixedStream object that shares a dictionary with a stream. ProxyMixedStream objects work exactly the same as MixedStream objects.
    
    **Parameters**
    
        **ID:** [str]
    
    """
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 16:21:56 2018

@author: yoelr
"""
from .. import Unit, MixedStream, PowerUtility
from math import pi, ceil
import copy
import numpy as np
from scipy.optimize import brentq, newton
from thermo import activity
from .designtools import vacuum_system, HNATable, FinalValue, \
                          VesselWeightAndWallThickness, Kvalue
from ._hx import HX, HXutility

exp = np.exp
ln = np.log

# USDA Biodiesel: 
#V = np.pi*D**2/4*Ht*0.3048**3
#if not (0.1 < V < 70):
#    raise DesignError(f"Volume is out of bounds for costing")
#lambda V, CE: CE*13*V**0.62 # V (m3)
__all__ = ('Flash', 'SplitFlash', 'PartitionFlash',
           'RatioFlash', 'Evaporator_PQin', 'Evaporator_PV')


# %% Flash solutions

flash_error = activity.Rachford_Rice_flash_error

def analytical_flash_solution(zs, Ks):
    """Solution for 2 or 3 component flash vessel."""
    # Not currently in use
    len_ = len(zs)
    if len_ == 2:
        z1, z2 = zs
        K1, K2 = Ks
        V_over_F = (-K1*z1 - K2*z2 + z1 + z2)/(K1*K2*z1 + K1*K2 *
                                               z2 - K1*z1 - K1*z2 - K2*z1 - K2*z2 + z1 + z2)
    elif len_ == 3:
        z1, z2, z3 = zs
        K1, K2, K3 = Ks
        V_over_F = (-K1*K2*z1/2 - K1*K2*z2/2 - K1*K3*z1/2 - K1*K3*z3/2 + K1*z1 + K1*z2/2 + K1*z3/2 - K2*K3*z2/2 - K2*K3*z3/2 + K2*z1/2 + K2*z2 + K2*z3/2 + K3*z1/2 + K3*z2/2 + K3*z3 - z1 - z2 - z3 - (K1**2*K2**2*z1**2 + 2*K1**2*K2**2*z1*z2 + K1**2*K2**2*z2**2 - 2*K1**2*K2*K3*z1**2 - 2*K1**2*K2*K3*z1*z2 - 2*K1**2*K2*K3*z1*z3 + 2*K1**2*K2*K3*z2*z3 - 2*K1**2*K2*z1*z2 + 2*K1**2*K2*z1*z3 - 2*K1**2*K2*z2**2 - 2*K1**2*K2*z2*z3 + K1**2*K3**2*z1**2 + 2*K1**2*K3**2*z1*z3 + K1**2*K3**2*z3**2 + 2*K1**2*K3*z1*z2 - 2*K1**2*K3*z1*z3 - 2*K1**2*K3*z2*z3 - 2*K1**2*K3*z3**2 + K1**2*z2**2 + 2*K1**2*z2*z3 + K1**2*z3**2 - 2*K1*K2**2*K3*z1*z2 + 2*K1*K2**2*K3*z1*z3 - 2*K1*K2**2*K3*z2**2 - 2*K1*K2**2*K3*z2*z3 - 2*K1*K2**2*z1**2 - 2*K1*K2**2*z1*z2 - 2*K1*K2**2*z1*z3 + 2*K1*K2**2*z2*z3 + 2*K1*K2*K3**2*z1*z2 - 2*K1*K2*K3**2*z1*z3 - 2*K1*K2*K3**2*z2*z3 - 2*K1*K2*K3**2*z3**2 + 4*K1*K2*K3*z1**2 + 4*K1*K2*K3*z1 *
                                                                                                                                                                                                       z2 + 4*K1*K2*K3*z1*z3 + 4*K1*K2*K3*z2**2 + 4*K1*K2*K3*z2*z3 + 4*K1*K2*K3*z3**2 + 2*K1*K2*z1*z2 - 2*K1*K2*z1*z3 - 2*K1*K2*z2*z3 - 2*K1*K2*z3**2 - 2*K1*K3**2*z1**2 - 2*K1*K3**2*z1*z2 - 2*K1*K3**2*z1*z3 + 2*K1*K3**2*z2*z3 - 2*K1*K3*z1*z2 + 2*K1*K3*z1*z3 - 2*K1*K3*z2**2 - 2*K1*K3*z2*z3 + K2**2*K3**2*z2**2 + 2*K2**2*K3**2*z2*z3 + K2**2*K3**2*z3**2 + 2*K2**2*K3*z1*z2 - 2*K2**2*K3*z1*z3 - 2*K2**2*K3*z2*z3 - 2*K2**2*K3*z3**2 + K2**2*z1**2 + 2*K2**2*z1*z3 + K2**2*z3**2 - 2*K2*K3**2*z1*z2 + 2*K2*K3**2*z1*z3 - 2*K2*K3**2*z2**2 - 2*K2*K3**2*z2*z3 - 2*K2*K3*z1**2 - 2*K2*K3*z1*z2 - 2*K2*K3*z1*z3 + 2*K2*K3*z2*z3 + K3**2*z1**2 + 2*K3**2*z1*z2 + K3**2*z2**2)**0.5/2)/(K1*K2*K3*z1 + K1*K2*K3*z2 + K1*K2*K3*z3 - K1*K2*z1 - K1*K2*z2 - K1*K2*z3 - K1*K3*z1 - K1*K3*z2 - K1*K3*z3 + K1*z1 + K1*z2 + K1*z3 - K2*K3*z1 - K2*K3*z2 - K2*K3*z3 + K2*z1 + K2*z2 + K2*z3 + K3*z1 + K3*z2 + K3*z3 - z1 - z2 - z3)
    return V_over_F

# %% Data
    
# Material density (lb∕ft^3)
rho_Mdict = {'Carbon steel': 490,
            'Low-alloy steel': None,
            'Stainless steel 304': 499.4,
            'Stainless steel 316': 499.4,
            'Carpenter 20CB-3': None,
            'Nickel-200': None,
            'Monel-400': None,
            'Inconel-600': None,
            'Incoloy-825': None,
            'Titanium': None}

# Vessel Material
F_Mdict = {'Carbon steel': 1.0,
           'Low-alloy steel': 1.2,
           'Stainless steel 304': 1.7,
           'Stainless steel 316': 2.1,
           'Carpenter 20CB-3': 3.2,
           'Nickel-200': 5.4,
           'Monel-400': 3.6,
           'Inconel-600': 3.9,
           'Incoloy-825': 3.7,
           'Titanium': 7.7}


# %% Flash

class Flash(Unit):
    """Create an equlibrium based flash drum with the option of having light non-keys and heavy non-keys completly separate into their respective phases. Design procedure is based on heuristics by Wayne D. Monnery & William Y. Svrcek [1]. The f.o.b cost is based on classical Guthrie's correlation [2]:

    :math:`C_{fob}^{2007} = 6500 V^{0.62} (0.1 < V < 70 m^3)`

    **Parameters**

        **P:** [float] Operating pressure (Pa)

        **User defines one:**
            * **Qin:** [float] Energy input (kJ/hr)
            * **V:** [float] Overall molar vapor fraction
            * **T:** [float] Operating temperature (K)
            * **x:** [array_like] Molar composition of liquid (for binary mixture)
            * **y:** [array_like] Molar composition of vapor (for binary mixture)     
    
        **Optional**
    
            **species_IDs:** tuple[str] IDs of equilibrium species
            
            **LNK**: tuple[str] Light non-keys
        
            **HNK**: tuple[str] Heavy non-keys

    **ins**
    
        [0] Input stream
        
    **outs**
    
        [0] Vapor product
        
        [1] Liquid product

    **References**

        [1] Design Two-Phase Separators Within the Right Limits, Chemical Engineering Progress Oct, 1993
    
        [2] I.K. Kookos, Preliminary Chemical Process Synthesis and Design, Tziolas Publishing, Thessalonika, Greece, 2008 (book in Greek).
    
    **Examples**
    
        :doc:`Flash Example`
    
    """
    _units = {'Vertical vessel weight': 'lb',
              'Horizontal vessel weight': 'lb',
              'Length': 'ft',
              'Diameter': 'ft',
              'Weight': 'lb',
              'Wall thickness': 'in'}
    _has_power_utility = False
    _N_heat_utilities = 0
    
    # Column material factor
    _F_Mstr = 'Carbon steel'
    _F_M = 1 
    
    #: If a vacuum system is needed, it will choose one according to this preference.
    vacuum_system_preference = 'Liquid-ring pump'
    
    #: [str] 'Horizontal', 'Vertical', or 'Default'
    SepType = 'Default'
    
    #: [float] Time it takes to raise liquid to half full (min)
    HoldupTime = 15  
    
    #: [float] Time it takes to reach from normal to maximum liquied level (min)
    SurgeTime = 7.5
    
    #: [bool] True if using a mist eliminator pad
    Mist = False
    
    _kwargs = {'species_IDs': None,  # Equilibrium species
               'LNK': None,   # light non-keys
               'HNK': None,   # heavy non-keys
               'V': None,   # Vapor fraction
               'T': None,   # Operating temperature
               'Qin': None, # Energy input
               'P': None,   # Operating Pressure
               'y': None,   # Vapor composition (of working species)
               'x': None}   # Liquid composition (of working species)

    _bounds = {'Vertical vessel weight': (4200, 1e6),
               'Horizontal vessel weight': (1e3, 9.2e5),
               'Diameter': (3, 21),
               'Vertical vessel Length': (12, 40)}

    @property
    def vessel_material(self):
        return self._F_Mstr
    @vessel_material.setter
    def vessel_material(self, vessel_material):
        try:
            self._F_M = F_Mdict[vessel_material]
        except KeyError:
            dummy = str(F_Mdict.keys())[11:-2]
            raise ValueError(f"Vessel material must be one of the following: {dummy}")
        self._F_Mstr = vessel_material  

    def _init(self):
        vap, liq = self.outs
        vap.phase = 'g'
        liq.phase = 'l'
        self._heat_exchanger = he = HXutility(None, outs=None) 
        self._heat_utilities = he._heat_utilities
        he._ins = self._ins
        self._mixedstream = ms = MixedStream(None)
        he._outs[0] = ms

    def _setup(self):
        self._has_hx = self._kwargs['Qin'] != 0
        if self._kwargs['P'] < 101325 and not self._power_utility:
            self._power_utility = PowerUtility()

    def _run(self):
        # Unpack
        vap, liq = self.outs
        feed = self.ins[0]

        # Vapor Liquid Equilibrium
        ms = self._mixedstream
        ms.empty()
        ms.liquid_mol = feed.mol
        ms.T = feed.T
        ms.VLE(**self._kwargs)

        # Set Values
        vap.mol = ms.vapor_mol
        liq.mol = ms.liquid_mol
        vap.T = liq.T = ms.T
        vap.P = liq.P = ms.P

    def _design(self):
        # Set horizontal or vertical vessel
        if self.SepType == 'Default':
            if self.outs[0].massnet/self.outs[1].massnet > 0.2:
                self.SepType = 'Vertical'
            else:
                self.SepType = 'Horizontal'

        # Run Vertical or Horizontal:
        if self.SepType == 'Vertical':
            out = self._vertical()
        elif self.SepType == 'Horizontal' or '(USDA) Biodiesel':
            out = self._horizontal()
        else:
            raise ValueError( f"SepType must be either 'Horizontal' or 'Vertical', not '{self.SepType}'")
        if self._has_hx: self._heat_exchanger._design()
        return out

    def _cost(self):
        Cost = self._results['Cost']
        Design = self._results['Design']
        W = Design['Weight']
        D = Design['Diameter']
        L = Design['Length']
        CE = self.CEPCI
        type_ = self.SepType
        
        # C_v: Vessel cost
        # C_pl: Platforms and ladders cost
        if type_ == 'Vertical':
            C_v = exp(7.0132 + 0.18255*ln(W) + 0.02297*ln(W)**2)
            C_pl = 361.8*D**0.7396*L**0.70684
        elif type_ == 'Horizontal':
            C_v = exp(8.9552 - 0.2330*ln(W) + 0.04333*ln(W)**2)
            C_pl = 2005*D**0.20294
        else:
            ValueError(f"SepType ({type_}) must be either 'Vertical', 'Horizontal' or 'Default'.")
            
        Cost['Vessel'] = CE/500*(C_v+C_pl)
        if self._has_hx:
            Cost['Heat exchanger'] = self._heat_exchanger._cost()['Heat exchanger']
        self._cost_vacuum()

    def _cost_vacuum(self):
        P = self._kwargs['P']
        if not P or P > 101320:
            return 
        
        r = self._results
        D = r['Design']
        C = r['Cost']
        vol = 0.02832 * np.pi * D['Length'] * (D['Diameter']/2)**2
        
        # If vapor is condensed, assume vacuum system is after condenser
        vapor = self.outs[0]
        hx = vapor.sink
        if isinstance(hx, HX):
            index = hx._ins.index(vapor)
            stream = hx._outs[index]
            if isinstance(stream, MixedStream):
                massflow = stream.vapor_massnet
                volflow = stream.vapor_volnet
            else:
                if stream.phase == 'g':
                    massflow = stream.massnet
                    volflow = stream.volnet
                else:
                    massflow = 0
                    volflow = 0
        else:
            massflow = vapor.massnet
            volflow = vapor.volnet
        
        power, cost = vacuum_system(massflow, volflow,
                                    P, vol,
                                    self.vacuum_system_preference)
        C['Liquid-ring pump'] = cost
        self._power_utility(power)

    def _design_parameters(self):
        # Retrieve run_args and properties
        vap, liq = self.outs[0:2]
        v = vap.quantity
        l = liq.quantity
        rhov = v('rho').to('lb/ft3').magnitude  # VapDensity
        rhol = l('rho').to('lb/ft3').magnitude  # LLiqDensity
        P = l('P').to('psi').magnitude  # Pressure

        # SepType (str), HoldupTime (min), SurgeTime (min), Mist (bool)
        SepType = self.SepType
        Th = self.HoldupTime
        Ts = self.SurgeTime
        Mist = self.Mist

        # Calculate the volumetric flowrate
        Qv = v('volnet').to('ft3/s').magnitude
        Qll = l('volnet').to('ft3/min').magnitude

        # Calculate Ut and set Uv
        K = Kvalue(P)

        # Adjust K value
        if not Mist and SepType == 'Vertical':
            K = K/2

        xs = liq.massfrac
        for s, x in zip(liq._species, xs):
            if x > 0.1:
                # TODO: Verify unifac groups to determine if glycol and/or amine
                dct = s.UNIFAC_groups
                keys = dct.keys()
                if 14 in keys:
                    OHs = dct[14]
                else:
                    OHs = 0
                if OHs > 1:
                    K = K*0.6
                elif any([i in keys for i in (28, 29, 30, 31, 32, 33)]):
                    K = K*0.8
                else:
                    OHs = 0

        Ut = K*((rhol - rhov) / rhov)**0.5
        Uv = 0.75*Ut
        # Calculate Holdup and Surge volume
        Vh = Th*Qll
        Vs = Ts*Qll
        return rhov, rhol, P, Th, Ts, Mist, Qv, Qll, Ut, Uv, Vh, Vs

    def _vertical(self):
        rhov, rhol, P, Th, Ts, Mist, Qv, Qll, Ut, Uv, Vh, Vs = self._design_parameters()

        # Calculate internal diameter, Dvd
        Dvd = (4.0*Qv/(pi*Uv))**0.5
        if Mist:
            D = FinalValue(Dvd + 0.4)
        else:
            D = FinalValue(Dvd)

        # Obtaining low liquid level height, Hlll
        Hlll = 0.5
        if P < 300:
            Hlll = 1.25

        # Calculate the height from Hlll to  Normal liq level, Hnll
        Hh = Vh/(pi/4.0*Dvd**2)
        if Hh < 1.0:
            Hh = 1.0
            Hh = FinalValue(Hh)

        # Calculate the height from Hnll to  High liq level, Hhll
        Hs = Vs/(pi/4.0*Dvd**2)
        if Hs < 0.5:
            Hs = 0.5
            Hs = FinalValue(Hs)

        # Calculate dN
        Qm = Qll + Qv
        lamda = Qll/Qm
        rhoM = rhol*lamda + rhov*(1-lamda)
        dN = (4*Qm/(pi*60.0/(rhoM**0.5)))**0.5
        dN = FinalValue(dN)

        # Calculate Hlin, assume with inlet diverter
        Hlin = 1.0 + dN

        # Calculate the vapor disengagement height
        Hv = 0.5*Dvd
        if Mist:
            Hv2 = 2.0 + dN/2.0
        else:
            Hv2 = 3.0 + dN/2.0

        if Hv2 < Hv:
            Hv = Hv2
        Hv = ceil(Hv)

        # Calculate total height, Ht
        Hme = 0.0
        if Mist:
            Hme = 1.5
        Ht = Hlll + Hh + Hs + Hlin + Hv + Hme
        Ht = FinalValue(Ht)

        # Check if LD is between 1.5 and 6.0
        converged = False
        LD = Ht/D
        while not converged:
            if (LD < 1.5):
                D = D - 0.5
                converged = False
            elif (LD > 6.0):
                D = D + 0.5
                converged = False
            else:
                converged = True
            LD = Ht/D

        # Calculate Vessel weight and wall thickness
        rho_M = rho_Mdict[self._F_Mstr]
        VW, VWT = VesselWeightAndWallThickness(P, D, Ht, rho_M)

        # Find maximum and normal liquid level
        # Hhll = Hs + Hh + Hlll
        # Hnll = Hh + Hlll

        # Results
        d = self._results['Design']
        self._checkbounds('Vertical vessel weight', VW)
        self._checkbounds('Vertical vessel length', Ht)
        d['SepType'] = 'Vertical'
        d['Length'] = Ht     # ft
        d['Diameter'] = D    # ft
        d['Weight'] = VW     # lb
        d['Wall thickness'] = VWT  # in
        return d
        
    def _horizontal(self):
        rhov, rhol, P, Th, Ts, Mist, Qv, Qll, Ut, Uv, Vh, Vs = self._design_parameters()
        maxIter = 50

        # Initialize LD
        if P > 0 and P <= 264.7:
            LD = 1.5/250.0*(P-14.7)+1.5
        elif P > 264.7 and P <= 514.7:
            LD = 1.0/250.0*(P-14.7)+2.0
        elif P > 514.7:
            LD = 5.0

        D = (4.0*(Vh+Vs)/(0.6*pi*LD))**(1.0/3.0)
        D = round(D)
        if D <= 4.0:
            D = 4.0

        outerIter = 0
        converged = False
        converged1 = False
        while not converged and outerIter < maxIter:
            outerIter += 1
            At = pi*(D**2)/4.0

            # Calculate Lower Liquid Area
            Hlll = round(0.5*D + 7.0)  
            Hlll = Hlll/12.0 # D is in ft but Hlll is in inches
            X = Hlll/D
            Y = HNATable(1, X)
            Alll = Y*At

            # Calculate the Vapor  disengagement area, Av
            Hv = 0.2*D
            if Mist and Hv <= 2.0:
                Hv = 2.0
            else:
                if Hv <= 1.0:
                    Hv = 1.0
            X = Hv/D
            Y = HNATable(1, X)
            Av = Y*At

            # Calculate minimum length for surge and holdup
            L = (Vh + Vs)/(At - Av - Alll)
            # Calculate liquid dropout
            Phi = Hv/Uv
            # Calculate actual vapor velocity
            Uva = Qv/Av
            # Calculate minimum length for vapor disengagement
            Lmin = Uva*Phi
            Li = L

            if L < Lmin:
                Li = Lmin

            if L < 0.8*Lmin:
                sign = 1.0
                needToIter = True
            elif L > 1.2*Lmin:
                if int(Mist) == 1 and Hv <= 2.0:
                    Hv = 2.0
                    Li = L
                elif not int(Mist) == 0 and Hv <= 1.0:
                    Hv = 1.0
                    Li = L
                else:
                    sign = -1.0
                    needToIter = True
            else:
                sign = -1.0
                needToIter = False

            if needToIter:
                innerIter = 0
                while not converged1 and innerIter < maxIter:
                    innerIter += 1
                    Hv = Hv + sign*0.5
                    if Mist and Hv <= 2.0:
                        Hv = 2.0
                    if Mist and Hv <= 1.0:
                        Hv = 1.0
                    X = Hv/D
                    Y = HNATable(1, X)
                    Av = Y*At

                    X = Hlll/D
                    Y = HNATable(1, X)
                    Alll = Y*At

                    Li = (Vh + Vs)/(At - Av - Alll)
                    Phi = Hv/Uv
                    Uva = Qv/Av
                    Lmin = Uva*Phi
                    if Li < 0.8*Lmin:
                        sign = 1.0
                        converged1 = False
                    elif Li > 1.2*Lmin:
                        sign = -1.0
                        converged1 = False
                    else:
                        Li = Li
                        converged1 = True
            
            L = Li
            LD = L/D
            # Check LD
            if LD < 1.2:
                if D <= 4.0:
                    D = D
                    converged = True
                else:
                    D = D - 0.5
                    converged = False

            if LD > 7.2:
                D = D + 0.5
                converged = False
            else:
                converged = True

        # Recalculate LD so it lies between 1.5 - 6.0
        converged = False
        while not converged:
            LD = L / D
            if (LD < 1.5) and D <= 4.0:
                L = L + 0.5
                converged = False
            elif LD < 1.5:
                D = D - 0.5
                converged = False
            elif (LD > 6.0):
                D = D + 0.5
                converged = False
            else:
                converged = True

        # Calculate vessel weight and wall thickness
        rho_M = rho_Mdict[self._F_Mstr]
        VW, VWT = VesselWeightAndWallThickness(P, D, L, rho_M)

        # # To check minimum Hv value
        # if int(Mist) == 1 and Hv <= 2.0:
        #     Hv = 2.0
        # if int(Mist) == 0 and Hv <= 1.0:
        #     Hv = 1.0

        # Calculate normal liquid level and High liquid level
        # Hhll = D - Hv
        # if (Hhll < 0.0):
        #     Hhll = 0.0
        # Anll = Alll + Vh/L
        # X = Anll/At
        # Y = HNATable(2, X)
        # Hnll = Y*D
        
        # Results
        d = self._results['Design']
        self._checkbounds('Horizontal vessel weight', VW)
        d['SepType'] = 'Horizontal'
        d['Length'] = L  # ft
        d['Diameter'] = D  # ft
        d['Weight'] = VW  # lb
        d['Wall thickness'] = VWT  # in
        return d
    
    

# %% Special


class SplitFlash(Flash):
    line = 'Flash'
    
    _kwargs = {'split': 1,  # component split fractions
               'T': 298.15,  # operating temperature (K)
               'P': 101325,
               'Qin': None}  # operating pressure (Pa)
    
    def _setup(self):
        top, bot = self.outs
        _, Tf, Pf, Qin = self._kwargs.values()
        self._has_hx = Qin != 0
        if self._kwargs['P'] < 101325 and not self._power_utility:
            self._power_utility = PowerUtility()
        bot.T = top.T = Tf
        bot.P = top.P = Pf

    def _run(self):
        split = self._kwargs['split']
        top, bot = self.outs
        net_mol = self._mol_in
        top.mol = net_mol*split
        bot.mol = net_mol - top.mol

    def _design(self):
        if self._has_hx:
            ms = MixedStream.sum(self._mixedstream, self.outs)
            self._heat_exchanger.outs = ms
        super()._design()

class PartitionFlash(Flash):
    """Create a PartitionFlash Unit that approximates outputs based on molar partition coefficients."""
    
    _kwargs = {'species_IDs': None,  # Partition species
               'Ks': None,           # Partition coefficients
               'LNK': None,          # light non-keys
               'HNK': None,          # heavy non-keys
               'P': None,            # Operating Pressure
               'T': None}            # Operating temperature
    
    _has_hx = True
    
    def _setup(self):
        top, bot = self.outs
        species_IDs, Ks, LNK, HNK, P, T = self._kwargs.values()
        index = top._IDs.index
        if P < 101325 and not self._power_utility:
            self._power_utility = PowerUtility()
        self._args = ([index(i) for i in species_IDs],
                      [index(i) for i in LNK],
                      [index(i) for i in HNK],
                       np.asarray(Ks), P, T)
        
    def _set_phases(self):
        top, bot = self.outs
        top.phase = 'g'
        bot.phase = 'l'
    
    def _run(self):
        feed = self.ins[0]
        ph1, ph2 = self.outs
        pos, LNK, HNK, Ks, P, T = self._args
        mol_in = feed.mol[pos]
        ph1.T = ph2.T = (T if T else feed.T)
        ph1.P = ph2.P = (P if P else feed.P)
        ph1.mol[LNK] = feed.mol[LNK]
        ph1.mol[HNK] = feed.mol[HNK]
        
        # Get zs and Ks to solve rashford rice equation
        molnet = sum(mol_in)
        zs = mol_in/molnet
        if hasattr(self, '_V'):
            self._V = V = newton(flash_error, self._V, args=(zs, Ks))
        elif len(zs) > 3:
            self._V = V = brentq(flash_error, 0, 1, (zs, Ks))
        else:
            V = analytical_flash_solution(zs, Ks)
        x = zs/(1 + V*(Ks-1))
        y = x*Ks
        ph1.mol[pos] = molnet*V*y
        ph2.mol[pos] = mol_in - ph1.mol[pos]
    
    _design = SplitFlash._design
    

class RatioFlash(Flash):
    _N_heat_utilities = 1
    _kwargs = {'Kspecies': None,  # list of species that correspond to Ks
               'Ks': None,  # array of molar ratio partition coefficinets,
               # Ks = y/x, where y and x are molar ratios of two different phases
               'top_solvents': [],  # list of species that correspond to top_split
               'top_split': [],  # list of splits for top_solvents
               'bot_solvents': [],  # list of species that correspond to bot_split
               'bot_split': []} # list of splits for bot_solvents  

    def _run(self):
        feed = self.ins[0]
        top, bot = self.outs
        kwargs = self._kwargs
        (Kspecies, Ks, top_solvents, top_split,
         bot_solvents, bot_split) = kwargs.values()
        sp_index = feed._IDs.index
        Kindex = [sp_index(s) for s in Kspecies]
        top_index = [sp_index(s) for s in top_solvents]
        bot_index = [sp_index(s) for s in bot_solvents]
        top_mol = top.mol; bot_mol = bot.mol; feed_mol = feed.mol
        top_mol[top_index] = feed_mol[top_index]*top_split
        bot_mol[top_index] = feed_mol[top_index]-top_mol[top_index]
        bot_mol[bot_index] = feed_mol[bot_index]*bot_split
        top_mol[bot_index] = feed_mol[bot_index]-bot_mol[bot_index]
        topnet = sum(top_mol[top_index])
        botnet = sum(bot_mol[bot_index])
        molnet = topnet+botnet
        top_mol[Kindex] = Ks * topnet * feed_mol[Kindex] / molnet  # solvent * mol ratio
        bot_mol[Kindex] = feed_mol[Kindex] - top_mol[Kindex]
        top.T, top.P = feed.T, feed.P
        bot.T, bot.P = feed.T, feed.P


# %% Single Component

class Evaporator_PQin(Unit):

    _kwargs = {'component': 'Water',  # ID of specie
               'Qin': 0,
               'P': 101325}

    def _setup(self):
        component, P = (self._kwargs[i] for i in ('component', 'P'))
        vapor, liquid = self.outs[:2]
        self._cached = cached = {}
        Tf = getattr(vapor._species, component).Tsat(P)

        # Set-up streams for energy balance
        vapor.empty()
        index = vapor._IDs.index(component)
        vapor.mol[index] = 1
        vapor.__dict__.update(phase='g', T=Tf, P=P)

        liquid.empty()
        liquid.mol[index] = 1
        liquid.__dict__.update(phase='l', T=Tf, P=P)

        no_ph_ch = type(liquid)(species=vapor.species)
        cached['vapor_H'] = vapor.H
        cached['liquid_H'] = liquid.H
        cached['no_ph_ch'] = no_ph_ch

    def _run(self):
        feed = self.ins[0]
        component, Qin = (self._kwargs[i] for i in ('component', 'Qin'))
        vapor, liquid = self.outs[:2]
        cached = self._cached

        # Set-up streams for energy balance
        vapor_H = cached['vapor_H']
        liquid_H = cached['liquid_H']
        no_ph_ch = cached['no_ph_ch']
        no_ph_ch.mol = copy.copy(feed.mol)
        index = no_ph_ch._IDs.index(component)
        no_ph_ch.mol[index] = 0

        no_ph_ch_H = no_ph_ch.H
        feed_H = feed.H
        # Optional if Qin also comes from condensing a side stream
        if len(self.ins) == 2:
            boiler_liq = self.outs[2]
            boiler_vap = self.ins[1]
            boiler_liq.copylike(boiler_vap)
            boiler_liq.phase = 'l'
            Qin = Qin + boiler_vap.H - boiler_liq.H

        # Energy balance to find vapor fraction
        def f(v):
            vapor_Hf = vapor_H*(v*feed.mol[index])
            liquid_Hf = liquid_H*((1-v)*feed.mol[index])
            return (vapor_Hf + liquid_Hf + no_ph_ch_H) - feed_H - Qin

        # Final vapor fraction
        v = newton(f, 0.5, tol=0.001)
        liquid.mol = no_ph_ch.mol
        if v < 0:
            v = 0
        elif v > 1:
            v = 1
        vapor.mol[index] = v*feed.mol[index]
        liquid.mol[index] = (1-v)*feed.mol[index]
        self._Qin = Qin
        self._V = v


class Evaporator_PV(Unit):
    _N_heat_utilities = 1
    _kwargs = {'component': 'Water',
               'V': 0.5,
               'P': 101325}

    def _setup(self):
        vapor, liquid = self.outs
        component, P = (self._kwargs[i] for i in ('component', 'P'))
        Tf = getattr(vapor._species, component).Tsat(P)
        vapor.__dict__.update(phase='g', T=Tf, P=P)
        liquid.__dict__.update(phase='l', T=Tf, P=P)

    def _run(self):
        feed = self.ins[0]
        component, V = (self._kwargs[i] for i in ('component', 'V'))
        vapor, liquid = self.outs
        index = vapor._IDs.index(component)
        vapor.mol[index] = V*feed.mol[index]
        liquid.mol = copy.copy(feed.mol)
        liquid.mol[index] = (1-V)*feed.mol[index]

        